from __future__ import annotations

from dataclasses import dataclass
from typing import ClassVar

from ..routing import AsyncViewType
from ..routing import HttpRoute as _HttpRoute
from ..routing import Router as _Router
from ..routing import Routes as _Routes
from ..routing import SocketRoute as _SocketRoute
from ..routing.extensions import MultimethodRoutes as _MultimethodRoutes
from .parameters import auto_params


@dataclass
class HttpRoute(_HttpRoute[AsyncViewType]):
    _auto_params: ClassVar = lambda cls, *args, **kwargs: auto_params(*args, **kwargs)


@dataclass
class SocketRoute(_SocketRoute[AsyncViewType]):
    _auto_params: ClassVar = lambda cls, *args, **kwargs: auto_params(*args, **kwargs)


class Routes(_Routes[AsyncViewType]):
    pass


class MultimethodRoutes(_MultimethodRoutes[AsyncViewType]):
    pass


class Router(_Router[AsyncViewType]):
    pass


__all__ = [
    "HttpRoute",
    "SocketRoute",
    "Routes",
    "MultimethodRoutes",
]



from __future__ import annotations

import asyncio
import json
from hashlib import md5

from ..openapi import OpenAPI as _OpenAPI
from .requests import request
from .responses import HTMLResponse, JSONResponse
from .routing import HttpRoute, Routes


class OpenAPI(_OpenAPI):
    @property
    def routes(self) -> Routes:
        async def redirect():
            return request.url.replace(path=request.url.path + "/")

        async def template():
            return HTMLResponse(self.html_template)

        async def json_docs():
            openapi = self.create_docs(request)
            return JSONResponse(
                openapi,
                headers={
                    "hash": md5(json.dumps(openapi).encode()).hexdigest(),
                    "reload": str(self.reload).lower(),
                },
            )

        async def heartbeat():
            async def g():
                openapi = self.create_docs(request)
                yield {
                    "id": md5(json.dumps(openapi).encode()).hexdigest(),
                    "data": json.dumps(openapi),
                }
                while not request.app.should_exit:
                    await asyncio.sleep(0.5)

            return g()

        return Routes(
            HttpRoute("", redirect, name=None),
            HttpRoute("/", template, name=None),
            HttpRoute("/json", json_docs, name=None),
            HttpRoute("/heartbeat", heartbeat, name=None),
        )



from pathlib import Path

import httpx
from baize.wsgi import Files, Response
from pydantic import BaseModel

from kui.wsgi import Kui


def test_pydantic_base_model():
    class Message(BaseModel):
        message: str

    app = Kui()

    @app.router.http("/message")
    def message():
        return Message(message="Hello, World!")

    @app.router.http("/{_:any}")
    def static_files():
        return Files(Path(__file__).absolute().parent, handle_404=Response(404))

    client = httpx.Client(
        base_url="http://testserver",
        transport=httpx.WSGITransport(app=app),  # type: ignore
    )
    response = client.get("/message")
    assert response.status_code == 200
    assert response.json() == {"message": "Hello, World!"}

    response = client.get("/test_response_convertors.py")
    assert response.status_code == 200



from __future__ import annotations

from decimal import Decimal
from uuid import UUID

import pytest

from kui.routing.tree import RadixTree


@pytest.fixture
def tree():
    tree = RadixTree()

    tree.append("/hello/{time:int}", ...)
    tree.append("/hello", ...)
    tree.append("/hello/world", ...)
    tree.append("/sayhi/{name}", ...)
    tree.append("/sayhi/{name}/suffix", ...)
    tree.append("/sayhi/{name}/avatar.{suffix}", ...)
    tree.append("/path/{filepath:any}", ...)
    tree.append("/decimal/{number:decimal}", ...)
    tree.append("/uuid/{id:uuid}", ...)

    return tree


@pytest.mark.parametrize(
    "path,params",
    [
        ("/hello", {}),
        ("/hello/world", {}),
        ("/hello/123", {"time": 123}),
        ("/sayhi/aber", {"name": "aber"}),
        ("/sayhi/aber/suffix", {"name": "aber"}),
        ("/sayhi/aber/avatar.png", {"name": "aber", "suffix": "png"}),
        ("/path/adsf", {"filepath": "adsf"}),
        ("/path/adsf/123", {"filepath": "adsf/123"}),
        ("/decimal/1.111", {"number": Decimal("1.111")}),
        (
            "/uuid/123e4567-e89b-12d3-a456-426655440000",
            {"id": UUID("123e4567-e89b-12d3-a456-426655440000")},
        ),
    ],
)
def test_tree_success_search(tree: RadixTree, path, params):
    result = tree.search(path)
    assert result is not None
    params, node = result
    assert params == params


@pytest.mark.parametrize(
    "path",
    ["", "/hello/", "/hello/world/", "/sayhi/aber/avatar"],
)
def test_tree_fail_search(tree: RadixTree, path):
    assert tree.search(path)[0] is None, f"error in {path}"


@pytest.mark.parametrize(
    "path",
    [
        "/path/{urlpath:any}/",
        "/sayhi/{name:int}/suffix",
        "/sayhi/{hi}/suffix",
        "/hello/{time:int}",
        "a",
        "/static/{filename}.py",
    ],
)
def test_tree_fail_add(tree, path):
    with pytest.raises(ValueError):
        tree.append(path, ...)


def test_tree_iterator(tree: RadixTree):
    for _0, _1 in zip(
        tree.iterator(),
        [
            ("/hello", ...),
            ("/hello/{time}", ...),
            ("/hello/world", ...),
            ("/sayhi/{name}", ...),
            ("/sayhi/{name}/suffix", ...),
            ("/sayhi/{name}/avatar.{suffix}", ...),
            ("/path/{filepath}", ...),
            ("/decimal/{number}", ...),
            ("/uuid/{id}", ...),
        ],
    ):
        assert _0 == _1


def test_static_preferred_over_dynamic_dynamic_first():
    tree: RadixTree[str] = RadixTree()

    tree.append("/users/{id}", "dynamic")
    tree.append("/users/me", "static")

    route, params = tree.search("/users/me")
    assert route is not None
    assert route[0] == "/users/me"
    assert route[2] == "static"
    assert params == {}


def test_static_preferred_over_dynamic_static_first():
    tree: RadixTree[str] = RadixTree()

    tree.append("/users/me", "static")
    tree.append("/users/{id}", "dynamic")

    route, params = tree.search("/users/me")
    assert route is not None
    assert route[0] == "/users/me"
    assert route[2] == "static"
    assert params == {}


def test_dynamic_still_matches_other_values():
    tree: RadixTree[str] = RadixTree()
    tree.append("/users/{id}", "dynamic")
    tree.append("/users/me", "static")

    route, params = tree.search("/users/42")
    assert route is not None
    assert route[0] == "/users/{id}"
    assert route[2] == "dynamic"
    assert params == {"id": "42"}


def test_static_preferred_on_sibling_branch():
    tree: RadixTree[str] = RadixTree()
    tree.append("/files/{name}", "dynamic_file")
    tree.append("/files/download", "static_download")

    route, params = tree.search("/files/download")
    assert route is not None
    assert route[0] == "/files/download"
    assert route[2] == "static_download"
    assert params == {}



from __future__ import annotations

import importlib
import os
from functools import reduce
from types import ModuleType
from typing import Any, Optional


def import_from_string(import_str: str) -> Any:
    module_str, _, attrs_str = import_str.partition(":")
    if not module_str or not attrs_str:
        raise ValueError(
            f'Import string "{import_str}" must be in format "<module>:<attribute>".'
        )

    return reduce(getattr, attrs_str.split("."), importlib.import_module(module_str))


def import_module(name: str) -> Optional[ModuleType]:
    """
    try importlib.import_module, nothing to do when module not be found.
    """
    if os.path.exists(os.path.join(os.getcwd(), name + ".py")) or os.path.exists(
        os.path.join(os.getcwd(), name, "__init__.py")
    ):
        return importlib.import_module(name)
    return None  # nothing to do when module not be found



from __future__ import annotations

from inspect import isfunction
from typing import TYPE_CHECKING, Any, Callable, List
from typing import cast as typing_cast

from ..routing import SyncViewType
from .requests import request
from .responses import HttpResponse


def required_method(method: str) -> Callable[[SyncViewType], SyncViewType]:
    """
    Set the acceptable request method of the function
    """
    allow_methods = {"HEAD", "GET"} if method == "GET" else {method}
    headers = {"Allow": ", ".join(allow_methods)}

    def decorator(function: SyncViewType) -> SyncViewType:
        if not isfunction(function):
            raise TypeError("`required_method` can only decorate function")

        def wrapper(*args, **kwargs):
            if request.method in allow_methods:
                return function(*args, **kwargs)
            elif request.method == "OPTIONS":
                return HttpResponse(headers=headers)
            else:
                return HttpResponse(status_code=405, headers=headers)

        setattr(wrapper, "__method__", method.upper())
        return typing_cast(SyncViewType, wrapper)

    return decorator


class HttpViewMeta(type):
    def __call__(self, *args: Any, **kwds: Any) -> Any:
        instance = super().__call__(*args, **kwds)
        return instance.__impl__()


class HttpView(metaclass=HttpViewMeta):
    HTTP_METHOD_NAMES = [
        "get",
        "post",
        "put",
        "patch",
        "delete",
        "head",
        "options",
        "trace",
    ]

    if TYPE_CHECKING:
        __methods__: List[str]

    def __init_subclass__(cls) -> None:
        cls.__methods__ = [m.upper() for m in cls.HTTP_METHOD_NAMES if hasattr(cls, m)]

    def __impl__(self) -> Any:
        handler = getattr(self, request.method.lower(), self.http_method_not_allowed)
        return handler()

    def http_method_not_allowed(self) -> HttpResponse:
        return HttpResponse(
            status_code=405, headers={"Allow": ", ".join(self.__methods__)}
        )

    @classmethod
    def options(cls) -> HttpResponse:
        """Handle responding to requests for the OPTIONS HTTP verb."""
        return HttpResponse(headers={"Allow": ", ".join(cls.__methods__)})



import httpx

from kui.wsgi import HttpView, Kui


def test_http_view():
    app = Kui()

    @app.router.http("/")
    class Home(HttpView):
        @classmethod
        def get(cls):
            return "OK"

    with httpx.Client(
        base_url="http://testServer",
        transport=httpx.WSGITransport(app=app),  # type: ignore
    ) as client:
        assert client.get("/").content == b"OK"

        assert client.post("/").status_code == 405

        assert client.options("/").headers["Allow"] == "GET, OPTIONS"



from kui.asgi import (
    FileResponse,
    HTMLResponse,
    JSONResponse,
    PlainTextResponse,
    RedirectResponse,
    SendEventResponse,
    StreamResponse,
)


def test_html_response():
    assert HTMLResponse[200] == {
        "200": {
            "content": {"text/html": {"schema": {"type": "string"}}},
            "description": "Request fulfilled, document follows",
            "headers": {},
        }
    }
    assert HTMLResponse[
        200,
        {
            "Location": {
                "description": "URL of created resource",
                "schema": {"type": "string"},
            }
        },
    ] == {
        "200": {
            "content": {"text/html": {"schema": {"type": "string"}}},
            "description": "Request fulfilled, document follows",
            "headers": {
                "Location": {
                    "description": "URL of created resource",
                    "schema": {"type": "string"},
                }
            },
        }
    }


def test_plain_text_response():
    assert PlainTextResponse[200] == {
        "200": {
            "content": {"text/plain": {"schema": {"type": "string"}}},
            "description": "Request fulfilled, document follows",
            "headers": {},
        }
    }
    assert PlainTextResponse[
        200,
        {
            "Location": {
                "description": "URL of created resource",
                "schema": {"type": "string"},
            }
        },
    ] == {
        "200": {
            "content": {"text/plain": {"schema": {"type": "string"}}},
            "description": "Request fulfilled, document follows",
            "headers": {
                "Location": {
                    "description": "URL of created resource",
                    "schema": {"type": "string"},
                }
            },
        }
    }


def test_json_response():
    assert JSONResponse[200] == {
        "200": {
            "description": "Request fulfilled, document follows",
            "headers": {},
        }
    }
    assert JSONResponse[
        200,
        {
            "Location": {
                "description": "URL of created resource",
                "schema": {"type": "string"},
            }
        },
    ] == {
        "200": {
            "description": "Request fulfilled, document follows",
            "headers": {
                "Location": {
                    "description": "URL of created resource",
                    "schema": {"type": "string"},
                }
            },
        }
    }


def test_redirect_response():
    assert RedirectResponse[200] == {
        "200": {
            "description": "Request fulfilled, document follows",
            "headers": {"Location": {"schema": {"type": "string"}}},
        }
    }
    assert RedirectResponse[
        200,
        {
            "Location": {
                "description": "URL of created resource",
                "schema": {"type": "string"},
            }
        },
    ] == {
        "200": {
            "description": "Request fulfilled, document follows",
            "headers": {
                "Location": {
                    "description": "URL of created resource",
                    "schema": {"type": "string"},
                }
            },
        }
    }


def test_file_response():
    assert FileResponse["text/html"] == {
        "200": {
            "description": "Request fulfilled, document follows",
            "content": {
                "text/html": {"schema": {"type": "string", "format": "binary"}}
            },
            "headers": {},
        },
        "206": {
            "description": "Partial content follows",
            "content": {
                "text/html": {"schema": {"type": "string", "format": "binary"}}
            },
            "headers": {},
        },
    }


def test_send_event_response():
    assert SendEventResponse[200] == {
        "200": {
            "description": "Request fulfilled, document follows",
            "content": {"text/event-stream": {"schema": {"type": "string"}}},
            "headers": {},
        }
    }


def test_stream_response():
    assert StreamResponse[200] == {
        "200": {
            "description": "Request fulfilled, document follows",
            "headers": {
                "Transfer-Encoding": {
                    "schema": {"type": "string"},
                    "description": "chunked",
                },
            },
        }
    }



def test_export_all():
    from kui.asgi import __all__ as asgi_all
    from kui.wsgi import __all__ as wsgi_all

    assert sorted(list(asgi_all)) == sorted(
        ["WebSocket", "websocket", "websocket_var", "SocketView", *wsgi_all]
    )



from __future__ import annotations

import json
import typing

from baize import asgi as baize_asgi
from baize.typing import ServerSentEvent

from ..responses import (
    FileResponseMixin,
    HTMLResponseMixin,
    JSONResponseMixin,
    PlainTextResponseMixin,
    RedirectResponseMixin,
    SendEventResponseMixin,
    StreamResponseMixin,
)
from .requests import request

__all__ = [
    "convert_response",
    "HttpResponse",
    "FileResponse",
    "HTMLResponse",
    "JSONResponse",
    "PlainTextResponse",
    "RedirectResponse",
    "SendEventResponse",
    "ServerSentEvent",
    "StreamResponse",
    "TemplateResponse",
]

HttpResponse = baize_asgi.Response


class JSONResponse(JSONResponseMixin, baize_asgi.JSONResponse):
    async def render(self, content: typing.Any) -> bytes:
        if not self.json_kwargs.get("default"):
            self.json_kwargs["default"] = request.app.json_encoder
        return json.dumps(content, **self.json_kwargs).encode(self.charset)


class FileResponse(
    FileResponseMixin,
    baize_asgi.FileResponse,
):
    pass


class PlainTextResponse(
    PlainTextResponseMixin,
    baize_asgi.PlainTextResponse,
):
    pass


class HTMLResponse(
    HTMLResponseMixin,
    baize_asgi.HTMLResponse,
):
    pass


class RedirectResponse(
    RedirectResponseMixin,
    baize_asgi.RedirectResponse,
):
    pass


class SendEventResponse(
    SendEventResponseMixin,
    baize_asgi.SendEventResponse,
):
    pass


class StreamResponse(
    StreamResponseMixin,
    baize_asgi.StreamResponse,
):
    pass


def TemplateResponse(
    name: str,
    context: typing.Mapping[str, typing.Any],
    status_code: int = 200,
    headers: typing.Optional[typing.Mapping[str, str]] = None,
    media_type: typing.Optional[str] = None,
    charset: typing.Optional[str] = None,
) -> HttpResponse:
    templates = request.app.templates
    if templates is None:
        raise RuntimeError(
            "You must assign a value to `app.templates` to use TemplateResponse"
        )

    return templates.TemplateResponse(
        name, context, status_code, headers, media_type, charset
    )


def convert_response(response: typing.Any) -> HttpResponse:
    """
    shortcut for convert response to HttpResponse

    Example:

        response = convert(response)

    It is equivalent to:

        if isinstance(response, tuple):
            response = app.response_converter(*response)
        else:
            response = app.response_converter(response)

    """
    if isinstance(response, tuple):
        return request.app.response_converter(*response)
    else:
        return request.app.response_converter(response)



import typing

SyncView = typing.Callable[..., typing.Any]
AsyncView = typing.Callable[..., typing.Awaitable[typing.Any]]


SyncViewType = typing.TypeVar("SyncViewType", bound=SyncView)
AsyncViewType = typing.TypeVar("AsyncViewType", bound=AsyncView)
ViewType = typing.TypeVar("ViewType", bound=typing.Union[SyncView, AsyncView])
MiddlewareType = typing.Callable[
    [typing.Callable[..., typing.Any]], typing.Callable[..., typing.Any]
]



import pytest

from kui.asgi.lifespan import asynccontextmanager_lifespan


@pytest.mark.asyncio
async def test_asynccontextmanager_lifespan(capsys):
    async def f(app):
        print("startup")
        yield
        print("shutdown")

    on_startup, on_shutdown = asynccontextmanager_lifespan(f)

    await on_startup(None)
    captured = capsys.readouterr()
    assert captured.out == "startup\n"

    await on_shutdown(None)
    captured = capsys.readouterr()
    assert captured.out == "shutdown\n"



from __future__ import annotations

import typing
from contextlib import ExitStack
from contextvars import ContextVar
from http import HTTPStatus

from baize.datastructures import URL, ContentType
from baize.exceptions import HTTPException
from baize.utils import cached_property
from baize.wsgi import HTTPConnection as BaiZeHTTPConnection
from baize.wsgi import Request as BaiZeRequest
from typing_extensions import Annotated

from .background import BackgroundTasks

if typing.TYPE_CHECKING:
    from .applications import Kui

from ..utils import State, bind_contextvar


class HTTPConnection(BaiZeHTTPConnection, typing.MutableMapping[str, typing.Any]):
    def __setitem__(self, name: str, value: typing.Any) -> None:
        self._environ[name] = value

    def __delitem__(self, name: str) -> None:
        del self._environ[name]

    @cached_property
    def state(self) -> State:
        return State(self.setdefault("state", {}))

    @cached_property
    def app(self) -> Kui:
        return self["app"]  # type: ignore

    def url_for(self, name: str, path_params: typing.Mapping[str, typing.Any]) -> URL:
        return self.url.replace(path=self.app.router.url_for(name, path_params))

    @property
    def background_tasks(self) -> BackgroundTasks:
        return self.state.setdefault("background_tasks", BackgroundTasks())

    @property
    def exit_stack(self) -> ExitStack:
        return self.state.setdefault("exit_stack", ExitStack())


class HttpRequest(BaiZeRequest, HTTPConnection):
    def data(
        self,
    ) -> Annotated[
        typing.Any,
        ContentType("application/json"),
        ContentType("application/x-www-form-urlencoded"),
        ContentType("multipart/form-data"),
    ]:
        content_type = self.content_type
        if content_type == "application/json":
            return self.json
        elif content_type in (
            "multipart/form-data",
            "application/x-www-form-urlencoded",
        ):
            return self.form

        raise HTTPException(HTTPStatus.UNSUPPORTED_MEDIA_TYPE)


http_connection_var: ContextVar[HTTPConnection] = ContextVar("http_connection")

http_connection = bind_contextvar(http_connection_var)

request_var: ContextVar[HttpRequest] = ContextVar("request")

request = bind_contextvar(request_var)



from __future__ import annotations

import pytest

from kui.openapi.extra_docs import merge_openapi_info


@pytest.mark.parametrize(
    "f,s,r",
    [
        ({"a": 1}, {"b": 1}, {"a": 1, "b": 1}),
        ({"a": {"a": 1}, "b": 1}, {"a": {"b": 1}}, {"a": {"a": 1, "b": 1}, "b": 1}),
        ({"a": (1, 2)}, {"a": (3,)}, {"a": [1, 2, 3]}),
    ],
)
def test_merge_openapi_info(f, s, r):
    assert merge_openapi_info(f, s) == r



from __future__ import annotations

import dataclasses
import inspect
import traceback
from contextlib import asynccontextmanager, nullcontext
from typing import TYPE_CHECKING, Any, AsyncGenerator, Callable, List, Tuple

from baize.typing import Receive, Scope, Send

if TYPE_CHECKING:
    from .applications import Kui


LifespanCallback = Callable[["Kui"], Any]


@dataclasses.dataclass
class Lifespan:
    on_startup: List[LifespanCallback] = dataclasses.field(default_factory=list)
    on_shutdown: List[LifespanCallback] = dataclasses.field(default_factory=list)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Handle ASGI lifespan messages, which allows us to manage application
        startup and shutdown events.
        """
        app: Kui = scope["app"]

        message = await receive()
        assert message["type"] == "lifespan.startup"
        try:
            for handler in self.on_startup:
                result = handler(app)
                if inspect.isawaitable(result):
                    await result
        except BaseException:
            msg = traceback.format_exc()
            await send({"type": "lifespan.startup.failed", "message": msg})
            raise
        await send({"type": "lifespan.startup.complete"})

        message = await receive()
        assert message["type"] == "lifespan.shutdown"
        try:
            for handler in self.on_shutdown:
                result = handler(app)
                if inspect.isawaitable(result):
                    await result
        except BaseException:
            msg = traceback.format_exc()
            await send({"type": "lifespan.shutdown.failed", "message": msg})
            raise
        await send({"type": "lifespan.shutdown.complete"})


def asynccontextmanager_lifespan(
    func: Callable[["Kui"], AsyncGenerator[Any, None]],
) -> Tuple[LifespanCallback, LifespanCallback]:
    """
    Convert `asynccontextmanager` function to `on_startup` and `on_shutdown`
    """
    context_manager_func = asynccontextmanager(func)
    context_manager: Any = nullcontext()

    async def on_startup(app: Kui) -> None:
        nonlocal context_manager
        context_manager = context_manager_func(app)
        await context_manager.__aenter__()

    async def on_shutdown(app: Kui) -> None:
        await context_manager.__aexit__(None, None, None)

    return on_startup, on_shutdown



from __future__ import annotations

from ..exceptions import HTTPException
from ..openapi.types import UploadFile
from ..parameters.field_functions import (
    Body,
    Cookie,
    Depends,
    Header,
    Path,
    Query,
)
from ..security import api_key_auth_dependency, basic_auth, bearer_auth
from .applications import FactoryClass, Kui
from .cors import allow_cors
from .openapi import OpenAPI
from .parameters import auto_params
from .requests import (
    HttpRequest,
    WebSocket,
    http_connection,
    http_connection_var,
    request,
    request_var,
    websocket,
    websocket_var,
)
from .responses import (
    FileResponse,
    HTMLResponse,
    HttpResponse,
    JSONResponse,
    PlainTextResponse,
    RedirectResponse,
    SendEventResponse,
    ServerSentEvent,
    StreamResponse,
    TemplateResponse,
    convert_response,
)
from .routing import HttpRoute, MultimethodRoutes, Routes, SocketRoute
from .templates import Jinja2Templates
from .views import HttpView, SocketView, required_method

__all__ = [
    "Kui",
    "FactoryClass",
    "OpenAPI",
    "HttpRequest",
    "WebSocket",
    "http_connection",
    "http_connection_var",
    "request",
    "request_var",
    "websocket",
    "websocket_var",
    "HTTPException",
    "Body",
    "Cookie",
    "Header",
    "Path",
    "Query",
    "Depends",
    "auto_params",
    "api_key_auth_dependency",
    "basic_auth",
    "bearer_auth",
    "UploadFile",
    "HttpResponse",
    "FileResponse",
    "HTMLResponse",
    "JSONResponse",
    "PlainTextResponse",
    "RedirectResponse",
    "ServerSentEvent",
    "SendEventResponse",
    "StreamResponse",
    "TemplateResponse",
    "convert_response",
    "required_method",
    "HttpView",
    "SocketView",
    "Routes",
    "MultimethodRoutes",
    "HttpRoute",
    "SocketRoute",
    "allow_cors",
    "Jinja2Templates",
]



import httpx
import pytest
from typing_extensions import Annotated

from kui.asgi import HttpRoute, HttpView, Kui, Path, Query, required_method
from kui.asgi import MultimethodRoutes as Routes
from kui.utils import safe_issubclass


def test_routes():
    async def endpoint():
        pass

    routes = Routes(base_class=HttpView)
    routes <<= Routes(
        HttpRoute("/login", endpoint),
        HttpRoute("/register", endpoint),
        base_class=HttpView,
    )
    assert routes == Routes(
        HttpRoute("/login", endpoint),
        HttpRoute("/register", endpoint),
        base_class=HttpView,
    )

    assert (
        routes
        << Routes(
            HttpRoute("/t/login", endpoint),
            HttpRoute("/t/register", endpoint),
            base_class=HttpView,
        )
    ) == (
        Routes(
            HttpRoute("/login", endpoint),
            HttpRoute("/register", endpoint),
            base_class=HttpView,
        )
        + Routes(
            HttpRoute("/t/login", endpoint),
            HttpRoute("/t/register", endpoint),
            base_class=HttpView,
        )
    )


def test_mulitmethodroutes():
    from kui.asgi import Kui

    async def list_user():
        pass

    routes = Routes(
        HttpRoute("/user", list_user) @ required_method("GET"), base_class=HttpView
    )

    @routes.http.post("/user")
    async def create_user():
        pass

    @routes.http.delete("/user")
    async def delete_user():
        pass

    app = Kui(routes=routes)

    endpoint = app.router.search("http", "/user")[1]
    assert safe_issubclass(endpoint, routes.base_class)
    assert hasattr(endpoint, "__methods__")
    assert (
        hasattr(endpoint, "get")
        and hasattr(endpoint, "post")
        and hasattr(endpoint, "delete")
    )


def test_mulitmethodroutes_with_prefix():
    from kui.asgi import Kui

    routes = Routes(base_class=HttpView)

    @routes.http.get("/user")
    async def list_user():
        pass

    @routes.http.post("/user")
    async def create_user():
        pass

    @routes.http.delete("/user")
    async def delete_user():
        pass

    app = Kui(routes="/api" // routes)

    endpoint = app.router.search("http", "/api/user")[1]
    assert safe_issubclass(endpoint, routes.base_class)
    assert hasattr(endpoint, "__methods__")
    assert (
        hasattr(endpoint, "get")
        and hasattr(endpoint, "post")
        and hasattr(endpoint, "delete")
    )


@pytest.mark.asyncio
async def test_mulitmethodroutes_with_parameters():
    routes = Routes(base_class=HttpView)

    @routes.http.get("/{name}", name=None)
    @routes.http.post("/{name}", name=None)
    @routes.http.put("/{name}", name=None)
    @routes.http.patch("/{name}", name=None)
    @routes.http.delete("/{name}", name=None)
    async def name(
        name: Annotated[str, Path()],
        status: Annotated[int, Query(200)],
    ):
        return name, status

    app = Kui(routes=routes)

    async with httpx.AsyncClient(
        transport=httpx.ASGITransport(app=app), base_url="http://testserver"
    ) as client:
        resp = await client.get("/aber")
        assert resp.text == "aber"

        resp = await client.post("/aber", params={"status": 201})
        assert resp.status_code == 201
        assert resp.text == "aber"

        resp = await client.get("/")
        assert resp.status_code == 404



from __future__ import annotations

import typing as t
from contextlib import contextmanager
from contextvars import ContextVar

__all__ = ["bind_contextvar"]

T = t.TypeVar("T")


def bind_contextvar(contextvar: ContextVar[T]) -> T:
    class ContextVarBind:
        __slots__ = ()

        def __getattr__(self, name):
            return getattr(contextvar.get(), name)

        def __setattr__(self, name, value):
            setattr(contextvar.get(), name, value)

        def __delattr__(self, name):
            delattr(contextvar.get(), name)

        def __getitem__(self, index):
            return contextvar.get()[index]  # type: ignore

        def __setitem__(self, index, value):
            contextvar.get()[index] = value  # type: ignore

        def __delitem__(self, index):
            del contextvar.get()[index]  # type: ignore

    return ContextVarBind()  # type: ignore


@contextmanager
def context_setter(var: ContextVar[T], value: T) -> t.Generator[None, None, None]:
    token = var.set(value)
    yield
    var.reset(token)



from __future__ import annotations

from typing import Any, Callable, Optional, TypeVar

from pydantic import Field
from typing_extensions import Annotated

from ..pydantic_compatible import Undefined
from .fields import Depends as DependInfo
from .fields import InBody, InCookie, InHeader, InPath, InQuery

T = TypeVar("T")


def Path(
    default: Any = Undefined,
    *,
    default_factory: Optional[Callable[[], Any]] = None,
    alias: str | None = None,
    title: str | None = None,
    description: str | None = None,
    **extra: Any,
) -> Any:
    """
    Used to provide extra information about a field.

    :param default: since this is replacing the field’s default, its first argument is used
      to set the default, use ellipsis (``...``) to indicate the field is required
    :param default_factory: callable that will be called when a default value is needed for this field
      If both `default` and `default_factory` are set, an error is raised.
    :param alias: the public name of the field
    :param title: can be any string, used in the schema
    :param description: can be any string, used in the schema
    :param **extra: any pydantic field kwargs
    """
    field_info = Field(
        default,
        default_factory=default_factory,
        alias=alias,
        title=title,
        description=description,
        **extra,
    )  # type: ignore
    return Annotated[Any, field_info, InPath()]


def Query(
    default: Any = Undefined,
    *,
    default_factory: Optional[Callable[[], Any]] = None,
    alias: str | None = None,
    title: str | None = None,
    description: str | None = None,
    **extra: Any,
) -> Any:
    """
    Used to provide extra information about a field.

    :param default: since this is replacing the field’s default, its first argument is used
      to set the default, use ellipsis (``...``) to indicate the field is required
    :param default_factory: callable that will be called when a default value is needed for this field
      If both `default` and `default_factory` are set, an error is raised.
    :param alias: the public name of the field
    :param title: can be any string, used in the schema
    :param description: can be any string, used in the schema
    :param **extra: any pydantic field kwargs
    """
    field_info = Field(
        default,
        default_factory=default_factory,
        alias=alias,
        title=title,
        description=description,
        **extra,
    )  # type: ignore
    return Annotated[Any, field_info, InQuery()]


def Header(
    default: Any = Undefined,
    *,
    default_factory: Optional[Callable[[], Any]] = None,
    alias: str | None = None,
    title: str | None = None,
    description: str | None = None,
    **extra: Any,
) -> Any:
    """
    Used to provide extra information about a field.

    :param default: since this is replacing the field’s default, its first argument is used
      to set the default, use ellipsis (``...``) to indicate the field is required
    :param default_factory: callable that will be called when a default value is needed for this field
      If both `default` and `default_factory` are set, an error is raised.
    :param alias: the public name of the field
    :param title: can be any string, used in the schema
    :param description: can be any string, used in the schema
    :param **extra: any pydantic field kwargs
    """
    field_info = Field(
        default,
        default_factory=default_factory,
        alias=alias.lower() if alias else None,
        title=title,
        description=description,
        **extra,
    )  # type: ignore
    return Annotated[Any, field_info, InHeader()]


def Cookie(
    default: Any = Undefined,
    *,
    default_factory: Optional[Callable[[], Any]] = None,
    alias: str | None = None,
    title: str | None = None,
    description: str | None = None,
    **extra: Any,
) -> Any:
    """
    Used to provide extra information about a field.

    :param default: since this is replacing the field’s default, its first argument is used
      to set the default, use ellipsis (``...``) to indicate the field is required
    :param default_factory: callable that will be called when a default value is needed for this field
      If both `default` and `default_factory` are set, an error is raised.
    :param alias: the public name of the field
    :param title: can be any string, used in the schema
    :param description: can be any string, used in the schema
    :param **extra: any pydantic field kwargs
    """
    field_info = Field(
        default,
        default_factory=default_factory,
        alias=alias,
        title=title,
        description=description,
        **extra,
    )  # type: ignore
    return Annotated[Any, field_info, InCookie()]


def Body(
    default: Any = Undefined,
    *,
    default_factory: Optional[Callable[[], Any]] = None,
    alias: str | None = None,
    title: str | None = None,
    description: str | None = None,
    exclusive: bool = False,
    **extra: Any,
) -> Any:
    """
    Used to provide extra information about a field.

    :param default: since this is replacing the field’s default, its first argument is used
      to set the default, use ellipsis (``...``) to indicate the field is required
    :param default_factory: callable that will be called when a default value is needed for this field
      If both `default` and `default_factory` are set, an error is raised.
    :param alias: the public name of the field
    :param title: can be any string, used in the schema
    :param description: can be any string, used in the schema
    :param exclusive: decide whether this field receives all parameters
    :param **extra: any pydantic field kwargs
    """
    field_info = Field(
        default,
        default_factory=default_factory,
        alias=alias,
        title=title,
        description=description,
        **extra,
    )  # type: ignore
    return Annotated[Any, field_info, InBody(exclusive=exclusive)]


def Depends(call: Callable, *, cache=True) -> Any:
    """
    Used to provide extra information about a field.

    :param call: callable that will be called when a dependency is needed for this field
    :param cache: whether to cache the result of the dependency call in the request state
    """
    return DependInfo(call, cache=cache)



import base64
from http import HTTPStatus
from typing import Callable, Tuple, Type, Union

from pydantic import Field
from typing_extensions import Annotated, Literal

from .exceptions import HTTPException
from .parameters.fields import InCookie, InHeader, InQuery

__all__ = [
    "bearer_auth",
]


def bearer_auth(
    authorization: Annotated[
        Union[str, None],
        Field(default=None, alias="authorization", title="Bearer Auth"),
        InHeader(
            security={
                "scheme": {"BearerAuth": {"type": "http", "scheme": "bearer"}},
                "required": {"BearerAuth": []},
            }
        ),
    ],
) -> Annotated[
    str,
    {
        401: {
            "description": HTTPStatus(401).description,
            "headers": {
                "WWW-Authenticate": {
                    "description": "Bearer token",
                    "schema": {"type": "string"},
                }
            },
        }
    },
]:
    """
    Bearer token authentication.
    """
    if authorization is None:
        raise HTTPException(
            401,
            headers={"WWW-Authenticate": "Bearer"},
        )
    try:
        type, token = authorization.strip().split(" ", maxsplit=1)
    except ValueError:
        raise HTTPException(
            401,
            headers={"WWW-Authenticate": "Bearer"},
        )
    if type != "Bearer":
        raise HTTPException(
            401,
            headers={"WWW-Authenticate": "Bearer"},
        )
    return token


def basic_auth(
    authorization: Annotated[
        Union[str, None],
        Field(default=None, alias="authorization", title="Basic Auth"),
        InHeader(
            security={
                "scheme": {"BasicAuth": {"type": "http", "scheme": "basic"}},
                "required": {"BasicAuth": []},
            }
        ),
    ],
) -> Annotated[
    Tuple[str, str],
    {
        401: {
            "description": HTTPStatus(401).description,
            "headers": {
                "WWW-Authenticate": {
                    "description": "Basic authentication",
                    "schema": {"type": "string"},
                }
            },
        }
    },
]:
    """
    Basic authentication.
    """
    if authorization is None:
        raise HTTPException(
            401,
            headers={"WWW-Authenticate": "Basic"},
        )
    try:
        type, token = authorization.strip().split(" ", maxsplit=1)
    except ValueError:
        raise HTTPException(
            401,
            headers={"WWW-Authenticate": "Basic"},
        )
    if type != "Basic":
        raise HTTPException(
            401,
            headers={"WWW-Authenticate": "Basic"},
        )
    username, password = base64.b64decode(token).decode("utf8").split(":")
    return username, password


def api_key_auth_dependency(
    name: str,
    position: Literal["query", "header", "cookie"] = "header",
) -> Callable[[Union[str, None]], str]:
    """
    Create API key authentication dependency.
    """
    class_: Union[Type[InQuery], Type[InHeader], Type[InCookie]]
    if position == "query":
        class_ = InQuery
    elif position == "header":
        class_ = InHeader
    elif position == "cookie":
        class_ = InCookie
    else:
        raise ValueError(
            f"Invalid position {position}, must be one of ('query', 'header', 'cookie')"
        )

    def api_key_auth(
        api_key: Annotated[
            Union[str, None],
            Field(
                default=None,
                alias=name,
                title="API Key",
            ),
            class_(
                security={
                    "scheme": {
                        "ApiKeyAuth": {"type": "apiKey", "name": name, "in": position}
                    },
                    "required": {"ApiKeyAuth": []},
                }
            ),
        ],
    ) -> Annotated[str, {401: {"description": HTTPStatus(401).description}}]:
        if api_key is None:
            raise HTTPException(401)
        return api_key

    return api_key_auth



from __future__ import annotations

import dataclasses
import functools
import sys
from pathlib import PurePath
from types import GeneratorType
from typing import Any, Callable, Iterable, List, Mapping, NoReturn, Optional, Type

from baize.datastructures import URL
from baize.typing import Environ, StartResponse
from baize.wsgi import Files, Hosts, Pages, Subpaths
from baize.wsgi import Router as BaizeRouter
from pydantic import BaseModel

from ..cors import CORSConfig
from ..responses import create_json_encoder
from ..routing import BaseRoute, MiddlewareType, NoMatchFound, SyncViewType
from ..utils import ImmutableAttribute, State
from ..utils.contextvars import context_setter
from .cors import allow_cors
from .exceptions import ErrorHandlerType, ExceptionMiddleware, HTTPException
from .requests import HttpRequest, http_connection_var, request_var
from .responses import (
    FileResponse,
    HttpResponse,
    JSONResponse,
    PlainTextResponse,
    RedirectResponse,
    SendEventResponse,
)
from .routing import Router
from .templates import BaseTemplates


@dataclasses.dataclass
class FactoryClass:
    http: Type[HttpRequest] = HttpRequest


class Kui:
    state: ImmutableAttribute[State] = ImmutableAttribute()

    def __init__(
        self,
        *,
        templates: Optional[BaseTemplates] = None,
        routes: Iterable[BaseRoute] = [],
        http_middlewares: List[MiddlewareType] = [],
        exception_handlers: Mapping[int | Type[BaseException], ErrorHandlerType] = {},
        cors_config: Optional[CORSConfig] = None,
        factory_class: FactoryClass = FactoryClass(),
        response_converters: Mapping[type, Callable[..., HttpResponse]] = {},
        json_encoder: Mapping[type, Callable[[Any], Any]] = {},
    ) -> None:
        self.should_exit = False

        self.state = State()
        self.response_converter = create_response_converter(response_converters)
        self.json_encoder = create_json_encoder(*json_encoder.items())
        self.should_exit = False
        self.factory_class = factory_class
        self.templates = templates

        http_middlewares = [*http_middlewares]

        self.exception_middleware = ExceptionMiddleware(exception_handlers)
        http_middlewares.append(self.exception_middleware)

        if cors_config is not None:
            http_middlewares.append(allow_cors(**cors_config))

        self.router = Router(routes, http_middlewares, [])

    def add_exception_handler(
        self, exc_class_or_status_code: int | Type[Exception], handler: ErrorHandlerType
    ) -> None:
        self.exception_middleware.add_exception_handler(
            exc_class_or_status_code, handler
        )

    def exception_handler(
        self, exc_class_or_status_code: int | Type[Exception]
    ) -> Callable[[ErrorHandlerType], ErrorHandlerType]:
        def decorator(func: ErrorHandlerType) -> ErrorHandlerType:
            self.add_exception_handler(exc_class_or_status_code, func)
            return func

        return decorator

    def app(self, environ: Environ, start_response: StartResponse) -> Iterable[bytes]:
        request = self.factory_class.http(environ)
        with (
            context_setter(http_connection_var, request),
            context_setter(request_var, request),
        ):
            try:
                try:
                    path_params, handler = self.router.search(
                        "http", request.get("PATH_INFO", "")
                    )
                    request["PATH_PARAMS"] = path_params
                    response = handler()
                except NoMatchFound:
                    http_exception: HTTPException[None] = HTTPException(404)
                    error_handler = self.exception_middleware.lookup_handler(
                        http_exception
                    )
                    if error_handler is None:
                        raise RuntimeError(
                            "No exception handler found for HTTPException(404)"
                        )
                    response = error_handler(http_exception)

                if isinstance(response, tuple):
                    response = self.response_converter(*response)
                else:
                    response = self.response_converter(response)

                yield from response(environ, start_response)
            finally:
                try:
                    request.exit_stack.__exit__(*sys.exc_info())
                    request.background_tasks()
                finally:
                    request.close()

    def __call__(
        self, environ: Environ, start_response: StartResponse
    ) -> Iterable[bytes]:
        environ["app"] = self

        return self.app(environ, start_response)


def create_response_converter(
    converters: Mapping[type, Callable[..., HttpResponse]],
) -> functools._SingleDispatchCallable[HttpResponse]:
    """
    Create a converter for convert response.
    """

    @functools.singledispatch
    def response_converter(*args: Any) -> HttpResponse:
        raise TypeError(
            f"Cannot find response_converter handler for this type: {type(args[0])}"
        )

    def _none(ret: Type[None]) -> NoReturn:
        raise TypeError(
            "Get 'None'. Maybe you need to add a return statement to the function."
        )

    response_converter.register(type(None), _none)
    response_converter.register(HttpResponse, lambda x: x)
    response_converter.register(Files, lambda x: x)
    response_converter.register(Pages, lambda x: x)
    response_converter.register(Subpaths, lambda x: x)
    response_converter.register(BaizeRouter, lambda x: x)
    response_converter.register(Hosts, lambda x: x)
    response_converter.register(dict, JSONResponse)
    response_converter.register(list, JSONResponse)
    response_converter.register(tuple, JSONResponse)
    response_converter.register(bytes, PlainTextResponse)
    response_converter.register(str, PlainTextResponse)
    response_converter.register(GeneratorType, SendEventResponse)
    response_converter.register(
        PurePath,
        lambda filepath, download_name=None: FileResponse(
            str(filepath), download_name=download_name
        ),
    )
    response_converter.register(URL, RedirectResponse)
    response_converter.register(BaseModel, JSONResponse)

    for type_, converter in converters.items():
        response_converter.register(type_, converter)

    return response_converter



import copy
from typing import Any, Dict, Tuple, Type

from pydantic import BaseModel
from pydantic import __version__ as pydantic_version

IS_V1 = pydantic_version.startswith("1.")

REF_TEMPLATE = "#/components/schemas/{model}"

__all__ = [
    "IS_V1",
    "validate_model",
    "get_model_fields",
    "get_model_json_schema",
    "create_root_model",
    "Undefined",
    "to_jsonable_python",
]

if IS_V1:
    from pydantic import create_model
    from pydantic.fields import ModelField, Undefined  # type: ignore
    from pydantic.json import pydantic_encoder as to_jsonable_python

    DEFINITIONS_KEY = "definitions"

    def validate_model(model: Type[BaseModel], v: Any) -> Tuple[Type[BaseModel], Any]:
        res = model.parse_obj(v)
        if hasattr(res, "__root__"):
            return model, res.__root__
        else:
            return model, res.__dict__

    def get_model_fields(model: Type[BaseModel]) -> Dict[str, ModelField]:  # type: ignore
        return model.__fields__  # type: ignore

    def get_model_json_schema(model: Type[BaseModel]) -> Dict[str, str]:
        return copy.deepcopy(model.schema(ref_template=REF_TEMPLATE))

    def create_root_model(type_: Any) -> Type[BaseModel]:
        return create_model("RootModel", __root__=(type_, ...))
else:
    from pydantic import RootModel
    from pydantic.fields import FieldInfo
    from pydantic_core import PydanticUndefined as Undefined
    from pydantic_core import to_jsonable_python

    DEFINITIONS_KEY = "$defs"

    def validate_model(model: Type[BaseModel], v: Any) -> Tuple[Type[BaseModel], Any]:
        res = model.model_validate(v)
        if isinstance(res, RootModel):
            return model, res.root
        else:
            return model, res.__dict__

    def get_model_fields(model: Type[BaseModel]) -> Dict[str, FieldInfo]:  # type: ignore
        return model.model_fields

    def get_model_json_schema(model: Type[BaseModel]) -> Dict[str, str]:
        return copy.deepcopy(model.model_json_schema(ref_template=REF_TEMPLATE))

    def create_root_model(type_: Any) -> Type[BaseModel]:
        return RootModel[type_]



import httpx
import pytest

from kui.asgi import HttpView, Kui, SocketView, websocket


@pytest.mark.asyncio
async def test_http_view():
    app = Kui()

    @app.router.http("/")
    class Home(HttpView):
        @classmethod
        async def get(cls):
            return "OK"

    async with httpx.AsyncClient(
        transport=httpx.ASGITransport(app=app), base_url="http://testserver"
    ) as client:
        response = await client.get("/")
        assert response.content == b"OK"

        response = await client.post("/")
        assert response.status_code == 405

        response = await client.options("/")
        assert response.headers["Allow"] == "GET, OPTIONS"


@pytest.mark.asyncio
async def test_socket_view():
    app = Kui()

    @app.router.websocket("/text")
    class Text(SocketView):
        encoding = "anystr"

        async def on_receive(self, data):
            await websocket.send_text(data)

    @app.router.websocket("/json")
    class Json(SocketView):
        encoding = "json"

        async def on_receive(self, data):
            await websocket.send_json(data)



from __future__ import annotations

import typing
from http import HTTPStatus

from typing_extensions import Annotated

from ..exceptions import ExceptionMiddlewareBase, HTTPException, RequestValidationError
from ..routing import SyncViewType
from .requests import request
from .responses import HttpResponse, JSONResponse, PlainTextResponse

__all__ = [
    "ExceptionMiddleware",
    "ErrorHandlerType",
    "HTTPException",
    "RequestValidationError",
]

ErrorHandlerType = typing.Callable[[typing.Any], typing.Any]


class ExceptionMiddleware(ExceptionMiddlewareBase[ErrorHandlerType]):
    def __call__(self, endpoint: SyncViewType) -> SyncViewType:
        def wrapper() -> typing.Any:
            try:
                return endpoint()
            except BaseException as exc:
                handler = self.lookup_handler(exc)
                if handler is None:
                    raise
                else:
                    return handler(exc)

        return wrapper  # type: ignore

    def _init_internal_handlers(self) -> None:
        self.add_exception_handler(HTTPException, self.http_exception)
        self.add_exception_handler(RequestValidationError, self.validation_error)

    @staticmethod
    def http_exception(exc: HTTPException) -> HttpResponse:
        if exc.status_code in {204, 304}:
            return HttpResponse(status_code=exc.status_code, headers=exc.headers)
        else:
            return request.app.response_converter(
                exc.content
                if exc.content is not None
                else HTTPStatus(exc.status_code).description,
                exc.status_code,
                exc.headers,
            )

    def validation_error(
        self, exc: RequestValidationError
    ) -> Annotated[
        HttpResponse, JSONResponse[422, {}, RequestValidationError.schema()]
    ]:
        if exc.in_ == "path":
            http_exception: HTTPException[None] = HTTPException(status_code=404)
            handler = self.lookup_handler(http_exception)
            if handler is None:
                raise RuntimeError("No exception handler found for HTTPException(404)")
            return handler(http_exception)
        else:
            return PlainTextResponse(
                exc.json(), status_code=422, media_type="application/json"
            )



from __future__ import annotations

import abc
import typing
from http import HTTPStatus

from pydantic import BaseModel

from .openapi import specification as spec
from .pydantic_compatible import create_root_model, to_jsonable_python
from .utils import safe_issubclass


def create_json_encoder(
    *json_encoders: typing.Tuple[type, typing.Callable[[typing.Any], typing.Any]],
) -> typing.Callable[[typing.Any], typing.Any]:
    def json_encoder(obj: typing.Any) -> typing.Any:
        for type_, encoder in json_encoders:
            if isinstance(obj, type_):
                return encoder(obj)
        return to_jsonable_python(obj)

    return json_encoder


class JSONResponseDocsMetaclass(abc.ABCMeta):
    def __getitem__(
        cls,
        parameters: typing.Tuple[
            int,
            typing.Dict[str, spec.Header | spec.Reference],
            typing.Type[BaseModel] | spec.Schema | typing.Any,
        ]
        | typing.Tuple[int, typing.Dict[str, spec.Header | spec.Reference]]
        | int,
    ) -> spec.Responses:
        """
        Use JSONResponse[status, headers, content] to describe response
        """
        status_code: int
        headers: typing.Dict[str, spec.Header | spec.Reference]
        content: typing.Type[BaseModel] | spec.Schema | typing.Any

        if isinstance(parameters, tuple):
            assert len(parameters) in (2, 3)
            # mypy can't check this
            if len(parameters) == 2:
                (status_code, headers), content = parameters, {}  # type: ignore
            else:
                status_code, headers, content = parameters  # type: ignore
        else:
            status_code, headers, content = parameters, {}, {}
        assert isinstance(status_code, int) or status_code == "default"

        docs: typing.Dict[str, typing.Any] = {
            str(status_code): {
                "description": HTTPStatus(status_code).description,
                "headers": headers,
            }
        }

        if content:
            real_content: (
                spec.Schema
                | typing.Type[BaseModel]
                | typing.Dict[typing.Any, typing.Any]
            )
            if isinstance(content, dict):
                real_content = content
            elif getattr(content, "__origin__", None) is None and safe_issubclass(
                content, BaseModel
            ):
                real_content = content
            else:
                real_content = create_root_model(content)
            docs[str(status_code)]["content"] = {
                "application/json": {"schema": real_content}
            }

        return docs


class JSONResponseMixin(metaclass=JSONResponseDocsMetaclass):
    """
    JSON response with OpenAPI docs support
    """


class FileResponseDocsMetaclass(abc.ABCMeta):
    def __getitem__(
        cls,
        parameters: typing.Tuple[str, typing.Dict[str, spec.Header | spec.Reference]]
        | str,
    ) -> spec.Responses:
        """
        Use FileResponse[content_type, headers] to describe response
        """
        if isinstance(parameters, tuple):
            content_type, headers = parameters
        else:
            content_type, headers = parameters, {}
        assert isinstance(content_type, str)

        return {
            "200": {
                "description": HTTPStatus.OK.description,
                "content": {
                    content_type: {"schema": {"type": "string", "format": "binary"}}
                },
                "headers": headers,
            },
            "206": {
                "description": HTTPStatus.PARTIAL_CONTENT.description,
                "content": {
                    content_type: {"schema": {"type": "string", "format": "binary"}}
                },
                "headers": headers,
            },
        }


class FileResponseMixin(metaclass=FileResponseDocsMetaclass):
    """
    File response with OpenAPI docs support
    """


class PlainTextResponseDocsMetaclass(abc.ABCMeta):
    def __getitem__(
        cls,
        parameters: typing.Tuple[int, typing.Dict[str, spec.Header | spec.Reference]]
        | int,
    ) -> spec.Responses:
        """
        Use PlainTextResponse[status, headers] to describe response
        """
        if isinstance(parameters, tuple):
            status_code, headers = parameters
        else:
            status_code, headers = parameters, {}
        assert isinstance(status_code, int)

        return {
            str(status_code): {
                "description": HTTPStatus(status_code).description,
                "content": {"text/plain": {"schema": {"type": "string"}}},
                "headers": headers,
            }
        }


class PlainTextResponseMixin(metaclass=PlainTextResponseDocsMetaclass):
    """
    Plain text response with OpenAPI docs support
    """


class HTMLResponseDocsMetaclass(abc.ABCMeta):
    def __getitem__(
        cls,
        parameters: typing.Tuple[int, typing.Dict[str, spec.Header | spec.Reference]]
        | int,
    ) -> spec.Responses:
        """
        Use HTMLResponse[status, headers] to describe response
        """
        if isinstance(parameters, tuple):
            status_code, headers = parameters
        else:
            status_code, headers = parameters, {}
        assert isinstance(status_code, int)

        return {
            str(status_code): {
                "description": HTTPStatus(status_code).description,
                "content": {"text/html": {"schema": {"type": "string"}}},
                "headers": headers,
            }
        }


class HTMLResponseMixin(metaclass=HTMLResponseDocsMetaclass):
    """
    HTML response with OpenAPI docs support
    """


class RedirectResponseDocsMetaclass(abc.ABCMeta):
    def __getitem__(
        cls,
        parameters: typing.Tuple[int, typing.Dict[str, spec.Header | spec.Reference]]
        | int,
    ) -> spec.Responses:
        """
        Use RedirectResponse[status, headers] to describe response
        """
        if isinstance(parameters, tuple):
            status_code, headers = parameters
        else:
            status_code, headers = parameters, {}
        assert isinstance(status_code, int)

        return {
            str(status_code): {
                "description": HTTPStatus(status_code).description,
                "headers": {
                    "Location": {"schema": {"type": "string"}},
                    **headers,
                },
            }
        }


class RedirectResponseMixin(metaclass=RedirectResponseDocsMetaclass):
    """
    Redirect response with OpenAPI docs support
    """


class SendEventResponseDocsMetaclass(abc.ABCMeta):
    def __getitem__(
        cls,
        parameters: typing.Tuple[int, typing.Dict[str, spec.Header | spec.Reference]]
        | int,
    ) -> spec.Responses:
        """
        Use SendEventResponse[status, headers] to describe response
        """
        if isinstance(parameters, tuple):
            status_code, headers = parameters
        else:
            status_code, headers = parameters, {}
        assert isinstance(status_code, int)

        # TODO: Follow the spec
        # https://github.com/OAI/OpenAPI-Specification/issues/396
        return {
            str(status_code): {
                "description": HTTPStatus(status_code).description,
                "content": {"text/event-stream": {"schema": {"type": "string"}}},
                "headers": headers,
            }
        }


class SendEventResponseMixin(metaclass=SendEventResponseDocsMetaclass):
    """
    Send event response with OpenAPI docs support
    """


class StreamResponseDocsMetaclass(abc.ABCMeta):
    def __getitem__(
        cls,
        parameters: typing.Tuple[int, typing.Dict[str, spec.Header | spec.Reference]]
        | int,
    ) -> spec.Responses:
        """
        Use StreamResponse[status, headers] to describe response
        """
        if isinstance(parameters, tuple):
            status_code, headers = parameters
        else:
            status_code, headers = parameters, {}
        assert isinstance(status_code, int)

        # TODO: Follow the spec
        # https://github.com/OAI/OpenAPI-Specification/issues/1576
        return {
            str(status_code): {
                "description": HTTPStatus(status_code).description,
                "headers": {
                    "Transfer-Encoding": {
                        "schema": {"type": "string"},
                        "description": "chunked",
                    },
                    **headers,
                },
            }
        }


class StreamResponseMixin(metaclass=StreamResponseDocsMetaclass):
    """
    Stream response with OpenAPI docs support
    """



from __future__ import annotations

import copy
import functools
import inspect
from itertools import groupby
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from typing import cast as typing_cast

from baize.datastructures import FormData
from pydantic import BaseModel, ValidationError, create_model
from pydantic.fields import FieldInfo
from typing_extensions import Annotated, Literal, get_args, get_origin, get_type_hints

from ..pydantic_compatible import (
    create_root_model,
    get_model_fields,
    get_model_json_schema,
    validate_model,
)

if TYPE_CHECKING:
    from ..asgi.requests import HTTPConnection as ASGIConnection
    from ..openapi import specification as spec
    from ..wsgi.requests import HTTPConnection as WSGIConnection

from ..exceptions import RequestValidationError
from ..utils import safe_issubclass
from .fields import (
    BaseHTTPFieldInfo,
    Depends,
    InBody,
    InCookie,
    InHeader,
    InPath,
    InQuery,
)

CallableObject = TypeVar("CallableObject", bound=Callable)


get_annotated_args = lambda tp: [
    j
    for i in (
        (get_annotated_args(t) if get_origin(t) is Annotated else [t])  # type: ignore
        for t in get_args(tp)
    )
    for j in i
]

sorted_groupby = lambda iterable, key: groupby(sorted(iterable, key=key), key=key)


def _merge_multi_value(
    items: Sequence[Tuple[str, Any]],
) -> Dict[str, Union[Any, List[Any]]]:
    """
    If there are values with the same key value, they are merged into a List.
    """
    return {
        k: v_list if len(v_list) > 1 else v_list[0]
        for k, v_list in (
            (k, [v for _, v in kv_iter])
            for k, kv_iter in sorted_groupby(items, lambda kv: kv[0])
        )
    }


def _parse_parameters_and_request_body_to_model(
    sig: inspect.Signature,
) -> Tuple[
    Dict[Literal["path", "query", "header", "cookie"], Type[BaseModel]] | None,
    Type[BaseModel] | None,
    Dict[Type[BaseModel], str],
    Dict[Literal["path", "query", "header", "cookie"], Dict[str, Any]],
]:
    raw_parameters: Dict[str, Any] = {
        key: {} for key in ["path", "query", "header", "cookie", "body"]
    }
    exclusive_models: Dict[Type[BaseModel], str] = {}
    security_info: Dict[
        Literal["path", "query", "header", "cookie"], Dict[str, Any]
    ] = {"path": {}, "query": {}, "header": {}, "cookie": {}}

    for name, param in sig.parameters.items():
        if not (
            get_origin(param.default) is Annotated
            or get_origin(param.annotation) is Annotated
        ):
            continue

        if param.POSITIONAL_ONLY:
            raise TypeError(
                f"Parameter {name} cannot be defined as positional only parameters."
            )

        annontated_define = param.annotation
        if get_origin(param.default) is Annotated:
            annontated_define = Annotated[annontated_define, param.default]

        type_, *annontated_list = get_annotated_args(annontated_define)
        kui_field: Union[InPath, InQuery, InHeader, InCookie, InBody]
        for kui_field in filter(
            lambda x: isinstance(x, (InPath, InQuery, InHeader, InCookie, InBody)),
            annontated_list,
        ):
            break
        else:
            # If there is no kui field, skip it.
            continue

        if kui_field.exclusive:
            model = create_root_model(type_)
            raw_parameters[kui_field._in] = model
            exclusive_models[model] = name
        else:
            if safe_issubclass(raw_parameters[kui_field._in], BaseModel):
                raise RuntimeError(
                    f"{kui_field._in.capitalize()}(exclusive=True) "
                    f"and {kui_field._in.capitalize()} cannot be used at the same time"
                )
            field_info = next(
                filter(lambda x: isinstance(x, FieldInfo), annontated_list)
            )
            raw_parameters[kui_field._in][name] = (type_, field_info)
            if (
                isinstance(kui_field, (InQuery, InHeader, InCookie))
                and kui_field.security
            ):
                security_info[kui_field._in][field_info.alias or name] = (
                    kui_field.security
                )

    for key, params in filter(
        lambda kv: kv[1],
        ((key, raw_parameters.pop(key)) for key in tuple(raw_parameters.keys())),
    ):
        if safe_issubclass(params, BaseModel):
            model = params
        else:
            model = create_model("temporary_model", **params)
        raw_parameters[key] = model

    request_body: Type[BaseModel] | None
    if "body" in raw_parameters:
        request_body = raw_parameters.pop("body")
    else:
        request_body = None

    parameters: Dict[str, Type[BaseModel]] | None
    if raw_parameters:
        parameters = typing_cast(Dict[str, Type[BaseModel]], raw_parameters)
    else:
        parameters = None

    return (
        typing_cast(
            Dict[Literal["path", "query", "header", "cookie"], Type[BaseModel]],
            parameters,
        ),
        request_body,
        exclusive_models,
        security_info,
    )


def _parse_depends_attrs(sig: inspect.Signature) -> Dict[str, Depends]:
    return {
        **{
            name: next(
                filter(lambda x: isinstance(x, Depends), get_args(param.annotation))
            )
            for name, param in sig.parameters.items()
            if (
                get_origin(param.annotation) is Annotated
                and any(isinstance(arg, Depends) for arg in get_args(param.annotation))
            )
        },
        **{
            name: param.default
            for name, param in sig.parameters.items()
            if isinstance(param.default, Depends)
        },
    }


def _create_new_signature(sig: inspect.Signature) -> inspect.Signature:
    return inspect.Signature(
        parameters=[
            param
            for param in sig.parameters.values()
            if not (
                isinstance(param.default, (BaseHTTPFieldInfo, Depends))
                or (
                    get_origin(param.annotation) is Annotated
                    and isinstance(get_args(param.annotation)[1], FieldInfo)
                )
            )
        ],
        return_annotation=sig.return_annotation,
    )


def _get_parameters_docs(
    m: Optional[Type[BaseModel]],
    position: Literal["path", "query", "header", "cookie"],
    security_fields: Set[str],
) -> List[spec.Parameter]:
    if m is None:
        return []

    _schemas: Dict[str, Any] = get_model_json_schema(m)
    properties: Dict[str, Any] = _schemas["properties"]
    required: Sequence[str] = _schemas.get("required", ())

    return [
        {
            "in": position,
            "name": name,
            "description": schema.pop("description", ""),
            "required": name in required,
            "schema": schema,
            "deprecated": schema.pop("deprecated", False),
        }
        for name, schema in properties.items()
        if name not in security_fields
    ]


def _merge_parameters_docs(
    x: List[spec.Parameter], y: List[spec.Parameter]
) -> List[spec.Parameter]:
    result = x + [
        param
        for param in y
        if not any(
            map(
                lambda x_param: (
                    x_param["name"] == param["name"] and x_param["in"] == param["in"]
                ),
                x,
            )
        )
    ]
    return result


def _get_response_docs(handler: Callable[..., Any]) -> List[Dict[Any, Any]]:
    response_docs: List[Dict[Any, Any]] = []
    for response in get_args(
        (
            get_type_hints(handler.__call__, include_extras=True)  # type: ignore
            if (
                callable(handler)
                and not (inspect.isfunction(handler) or inspect.ismethod(handler))
            )
            else get_type_hints(handler, include_extras=True)
        ).get("return")
    ):
        if isinstance(response, dict):
            response_docs.append(response)
    return copy.deepcopy(response_docs)


def _update_docs(
    old_handler: Callable[..., Any],
    handler: Callable[..., Any],
    parameters: Dict[Literal["path", "query", "header", "cookie"], Type[BaseModel]]
    | None,
    request_body: Type[BaseModel] | None,
    depend_functions: Dict[str, Callable[..., Any]],
    security_info: Dict[Literal["path", "query", "header", "cookie"], Dict[str, Any]],
) -> None:
    if inspect.ismethod(handler):
        handler = handler.__func__  # type: ignore

    if isinstance(handler.__doc__, str):
        clean_doc = inspect.cleandoc(handler.__doc__)
        if not hasattr(handler, "__docs_summary__") and not hasattr(
            handler, "__docs_description__"
        ):
            for k, value in zip(("summary", "description"), clean_doc.split("\n\n", 1)):
                setattr(handler, f"__docs_{k}__", value)
        elif not hasattr(handler, "__docs_description__"):
            setattr(handler, "__docs_description__", clean_doc)

    if request_body is not None:
        __request_body__: List[Type[BaseModel]] = getattr(
            handler, "__docs_request_body__", []
        )
        __request_body__.append(request_body)
        setattr(handler, "__docs_request_body__", __request_body__)

    if parameters is not None:
        __parameters__: List[Any] = getattr(handler, "__docs_parameters__", [])
        __security__: List[Any] = getattr(handler, "__docs_security__", [])
        for position, model in parameters.items():
            __parameters__ = _merge_parameters_docs(
                __parameters__,
                _get_parameters_docs(
                    model, position, set(security_info[position].keys())
                ),
            )
        __security__.extend(info for p in security_info.values() for info in p.values())
        setattr(handler, "__docs_security__", __security__)
        setattr(handler, "__docs_parameters__", __parameters__)

    setattr(handler, "__docs_responses__", parse_docs_responses(old_handler))

    for func in depend_functions.values():
        __request_body__ = getattr(handler, "__docs_request_body__", [])
        __request_body__.extend(getattr(func, "__docs_request_body__", []))
        setattr(handler, "__docs_request_body__", __request_body__)

        setattr(
            handler,
            "__docs_parameters__",
            _merge_parameters_docs(
                getattr(handler, "__docs_parameters__", []),
                getattr(func, "__docs_parameters__", []),
            ),
        )

        __security__ = getattr(handler, "__docs_security__", [])
        __security__.extend(getattr(func, "__docs_security__", []))
        setattr(handler, "__docs_security__", __security__)

        __responses__: List[Dict[Any, Any]] = getattr(handler, "__docs_responses__", [])
        __responses__.extend(_get_response_docs(func))
        setattr(handler, "__docs_responses__", __responses__)


def _validate_parameters_and_request_body(
    parameters: Dict[Literal["path", "query", "header", "cookie"], Type[BaseModel]],
    request_body: Type[BaseModel] | None,
    request: ASGIConnection | WSGIConnection,
) -> Generator[None, Any, List[Tuple[Type[BaseModel], Any]]]:
    data = []

    if "path" in parameters:
        try:
            data.append(validate_model(parameters["path"], request.path_params))
        except ValidationError as e:
            raise RequestValidationError(e, "path")

    if "query" in parameters:
        try:
            data.append(
                validate_model(
                    parameters["query"],
                    _merge_multi_value(request.query_params.multi_items()),
                )
            )
        except ValidationError as e:
            raise RequestValidationError(e, "query")

    if "header" in parameters:
        try:
            data.append(validate_model(parameters["header"], request.headers._dict))
        except ValidationError as e:
            raise RequestValidationError(e, "header")

    if "cookie" in parameters:
        try:
            data.append(validate_model(parameters["cookie"], request.cookies))
        except ValidationError as e:
            raise RequestValidationError(e, "cookie")

    # try to get body model and parse
    if request_body:
        _body_data = yield
        if isinstance(_body_data, FormData):
            _body_data = _merge_multi_value(_body_data.multi_items())

        try:
            data.append(validate_model(request_body, _body_data))
        except ValidationError as e:
            raise RequestValidationError(e, "body")

    return data


def _convert_model_data_to_keyword_arguments(
    data: List[Tuple[Type[BaseModel], Any]],
    exclusive_models: Dict[Type[BaseModel], str],
) -> Dict[str, Any]:
    result = {}
    for model, value in data:
        if model.__name__ == "temporary_model":
            result.update(value)
        else:
            result[exclusive_models[model]] = value
    return result


def _create_new_class(cls):
    """
    Create a fake class for auto-bound parameters.
    """

    @functools.wraps(cls, updated=())
    class NewClass(cls):
        pass

    return NewClass


def create_auto_params(
    create_new_callback: Callable[[CallableObject], CallableObject],
) -> Callable[[CallableObject], CallableObject]:
    """
    Create auto_params
    """

    def auto_params(handler: CallableObject) -> CallableObject:
        if hasattr(handler, "__methods__"):
            new_class = _create_new_class(handler)
            for method in map(lambda x: x.lower(), handler.__methods__):
                old_callback = getattr(handler, method)
                new_callback = create_new_callback(old_callback)
                setattr(new_class, method, new_callback)  # note: set to new class
            setattr(
                new_class,
                "__raw_handler__",
                getattr(handler, "__raw_handler__", handler),
            )
            return new_class  # type: ignore
        else:
            old_callback = handler
            new_callback = create_new_callback(old_callback)
            setattr(
                new_callback,
                "__raw_handler__",
                getattr(handler, "__raw_handler__", handler),
            )
            return new_callback

    return auto_params


def update_wrapper(
    new_handler: CallableObject, old_handler: Callable
) -> CallableObject:
    """
    Update wrapper for auto-bound parameters.
    """
    for attr in dir(old_handler):
        if attr.startswith("__docs_") or attr in ("__method__", "__methods__"):
            setattr(new_handler, attr, getattr(old_handler, attr))

    setattr(new_handler, "__raw_handler__", old_handler)

    return new_handler


def parse_docs_responses(callback):
    return [
        *getattr(callback, "__docs_responses__", []),
        *_get_response_docs(callback),
    ]



from __future__ import annotations

import abc
from typing import Any, Callable, Dict, Generic, List, Mapping, Type, TypeVar, Union

from baize.exceptions import HTTPException
from pydantic import ValidationError
from typing_extensions import Literal


class RequestValidationError(Exception):
    def __init__(
        self,
        validation_error: ValidationError,
        in_: Literal["path", "query", "header", "cookie", "body"],
    ) -> None:
        self.validation_error = validation_error
        self.in_ = in_

    def errors(self) -> List[Dict[str, Any]]:
        errors = self.validation_error.errors()
        for error in errors:
            error["in"] = self.in_  # type: ignore
        return errors  # type: ignore

    def json(self, *, indent: Union[None, int] = 2) -> str:
        return self.validation_error.json(indent=indent)

    @staticmethod
    def schema() -> Dict[str, Any]:
        return {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "loc": {
                        "title": "Location",
                        "description": "error field",
                        "type": "array",
                        "items": {"type": "string"},
                    },
                    "type": {
                        "title": "Type",
                        "description": "error type",
                        "type": "string",
                    },
                    "msg": {
                        "title": "Message",
                        "description": "error message",
                        "type": "string",
                    },
                    "ctx": {
                        "title": "Context",
                        "description": "error context",
                        "type": "string",
                    },
                    "in": {
                        "title": "In",
                        "type": "string",
                        "enum": ["path", "query", "header", "cookie", "body"],
                    },
                },
                "required": ["loc", "type", "msg"],
            },
        }


ErrorHandlerType = TypeVar("ErrorHandlerType", bound=Callable)


class ExceptionMiddlewareBase(Generic[ErrorHandlerType], abc.ABC):
    def __init__(
        self,
        handlers: Mapping[Union[int, Type[BaseException]], ErrorHandlerType] = {},
    ) -> None:
        self._status_handlers: Dict[int, ErrorHandlerType] = {}
        self._exception_handlers: Dict[Type[BaseException], ErrorHandlerType] = {}
        self._init_internal_handlers()
        for key, value in handlers.items():
            self.add_exception_handler(key, value)

    @abc.abstractmethod
    def _init_internal_handlers(self) -> None:
        raise NotImplementedError

    def add_exception_handler(
        self,
        exc_class_or_status_code: Union[int, Type[BaseException]],
        handler: ErrorHandlerType,
    ) -> None:
        if isinstance(exc_class_or_status_code, int):
            self._status_handlers[exc_class_or_status_code] = handler
        else:
            self._exception_handlers[exc_class_or_status_code] = handler

    def lookup_handler(self, exc: BaseException) -> ErrorHandlerType | None:
        handler = None
        if isinstance(exc, HTTPException):
            handler = self._status_handlers.get(exc.status_code)
        if handler is None:
            handler = self._lookup_exception_handler(type(exc))
        return handler

    def _lookup_exception_handler(
        self, exc_type: Type[BaseException]
    ) -> ErrorHandlerType | None:
        for cls in exc_type.__mro__:
            if cls in self._exception_handlers:
                return self._exception_handlers[cls]
        return None



import inspect
import io

import httpx
import pytest
from httpx import Client
from pydantic import BaseModel
from typing_extensions import Annotated

from kui.wsgi import (
    Body,
    Cookie,
    Depends,
    Header,
    Kui,
    Path,
    Query,
    UploadFile,
    auto_params,
    request,
)


def test_path():
    app = Kui()

    @app.router.http.get("/")
    @app.router.http.get("/{name}", name=None)
    @app.router.http.get("/{id:int}", name=None)
    def path(name: Annotated[str, Path()]):
        return name

    with Client(
        base_url="http://testServer",
        transport=httpx.WSGITransport(app=app),  # type: ignore
    ) as client:
        resp = client.get("/aber")
        assert resp.text == "aber"

        resp = client.get("/")
        assert resp.status_code == 404

        resp = client.get("/1")
        assert resp.status_code == 404

    assert not inspect.signature(app.router.search("http", "/aber")[1]).parameters


def test_query():
    app = Kui()

    @app.router.http.get("/")
    def query(name: Annotated[str, Query(...)]):
        return name

    with Client(
        base_url="http://testServer",
        transport=httpx.WSGITransport(app=app),  # type: ignore
    ) as client:
        resp = client.get("/", params={"name": "aber"})
        assert resp.text == "aber"

        resp = client.get("/", params={})
        assert resp.status_code == 422

    assert not inspect.signature(app.router.search("http", "/")[1]).parameters


def test_header():
    app = Kui()

    @app.router.http.get("/")
    def header(name: Annotated[str, Header(alias="Name")]):
        return name

    with Client(
        base_url="http://testServer",
        transport=httpx.WSGITransport(app=app),  # type: ignore
    ) as client:
        resp = client.get("/", headers={"name": "aber"})
        assert resp.text == "aber"

        resp = client.get("/", headers={"name0": "aber"})
        assert resp.status_code == 422

    assert not inspect.signature(app.router.search("http", "/")[1]).parameters


def test_cookie():
    app = Kui()

    @app.router.http.get("/")
    def cookie(name: Annotated[str, Cookie()]):
        return name

    with Client(
        base_url="http://testServer",
        transport=httpx.WSGITransport(app=app),  # type: ignore
        cookies={"name": "aber"},
    ) as client:
        resp = client.get("/")
        assert resp.text == "aber"

    with Client(
        base_url="http://testServer",
        transport=httpx.WSGITransport(app=app),  # type: ignore
        cookies={"name0": "aber"},
    ) as client:
        resp = client.get("/")
        assert resp.status_code == 422

    assert not inspect.signature(app.router.search("http", "/")[1]).parameters


def test_body():
    app = Kui()

    @app.router.http.post("/")
    def body(name: Annotated[str, Body()]):
        return name

    with Client(
        base_url="http://testServer",
        transport=httpx.WSGITransport(app=app),  # type: ignore
    ) as client:
        resp = client.post("/", data={"name": "aber"})
        assert resp.text == "aber"

        resp = client.post("/", data={"name0": "aber"})
        assert resp.status_code == 422

    assert not inspect.signature(app.router.search("http", "/")[1]).parameters


def test_depend():
    app = Kui()

    def get_name(name: Annotated[str, Body(...)]):
        return name

    @app.router.http.post("/")
    def homepage(name: Annotated[str, Depends(get_name)]):
        return name

    in_gen = False

    def gen(name: Annotated[str, Query(...)]):
        nonlocal in_gen
        in_gen = True
        try:
            yield name
        finally:
            in_gen = False

    @app.router.http.post("/gen")
    def depend_gen(name: Annotated[str, Depends(gen)]):
        assert in_gen
        return name

    def async_get_name(name: Annotated[str, Body(...)]):
        return name

    @app.router.http.post("/async/")
    def depend_async(name: Annotated[str, Depends(async_get_name)]):
        return name

    def async_gen(name: Annotated[str, Query(...)]):
        nonlocal in_gen
        in_gen = True
        try:
            yield name
        finally:
            in_gen = False

    @app.router.http.post("/async/gen")
    def depend_async_gen(name: Annotated[str, Depends(async_gen)]):
        assert in_gen
        return name

    count = 0

    def cached_get_name(name: Annotated[str, Body(...)]):
        nonlocal count
        count += 1
        return name

    @app.router.http.post("/cache/enable")
    def cache_enable(
        name0: Annotated[str, Depends(cached_get_name, cache=True)],
        name1: Annotated[str, Depends(cached_get_name, cache=True)],
    ):
        assert name0 == name1
        return name0

    @app.router.http.post("/cache/disable")
    def cache_disable(
        name0: Annotated[str, Depends(cached_get_name, cache=False)],
        name1: Annotated[str, Depends(cached_get_name, cache=False)],
    ):
        assert name0 == name1
        return name0

    with Client(
        base_url="http://testServer",
        transport=httpx.WSGITransport(app=app),  # type: ignore
    ) as client:
        resp = client.post("/", json={"name": "aber"})
        assert resp.text == "aber"

        resp = client.post("/gen", params={"name": "123"})
        assert resp.text == "123"
        assert not in_gen

        resp = client.post("/async/", json={"name": "123"})
        assert resp.text == "123"

        resp = client.post("/async/gen", params={"name": "123"})
        assert resp.text == "123"
        assert not in_gen

        resp = client.post("/cache/enable", json={"name": "123"})
        assert resp.text == "123"
        assert count == 1

        resp = client.post("/cache/disable", json={"name": "123"})
        assert resp.text == "123"
        assert count == 3


def test_middleware():
    app = Kui()

    def middleware(endpoint):
        def middleware_wrapper(query: Annotated[str, Query(...)]):
            return endpoint()

        return middleware_wrapper

    @app.router.http.get("/", middlewares=[middleware])
    def cookie(name: Annotated[str, Cookie()]):
        return name

    with Client(
        base_url="http://testServer",
        transport=httpx.WSGITransport(app=app),  # type: ignore
        cookies={"name": "aber"},
    ) as client:
        resp = client.get("/")
        assert resp.status_code == 422

        resp = client.get("/?query=123")
        assert resp.text == "aber"

    assert not inspect.signature(app.router.search("http", "/")[1]).parameters


def test_upload_file():
    app = Kui()

    @app.router.http.post("/")
    def upload_file(file: Annotated[UploadFile, Body(...)]):
        return {
            "filename": file.filename,
            "content": file.read().decode("utf8"),
        }

    with httpx.Client(
        base_url="http://testserver",
        transport=httpx.WSGITransport(app=app),  # type: ignore
    ) as client:
        resp = client.post("/", files={"file": ("file", io.BytesIO(b"123"))})
        assert resp.json() == {"filename": "file", "content": "123"}

        resp = client.post("/", files={"file0": io.BytesIO(b"123")})
        assert resp.status_code == 422


def test_parameters_nest_model():
    app = Kui()

    class User(BaseModel):
        name: str
        age: int

    @app.router.http.post("/")
    def create_user(user: Annotated[User, Body(...)]):
        assert user.age >= 18
        return user

    with httpx.Client(
        base_url="http://testserver",
        transport=httpx.WSGITransport(app=app),  # type: ignore
    ) as client:
        resp = client.post("/", json={"user": {"name": "aber", "age": 18}})
        assert resp.json() == {"name": "aber", "age": 18}

        resp = client.post("/", json={"name": "aber", "age": 18})
        assert resp.status_code == 422


def test_auto_params():
    app = Kui()

    @app.router.http.get("/")
    def index(name: Annotated[str, Query(...)]):
        return tf()

    @auto_params
    def tf(name: str = Query(...)):
        return name

    @app.router.http("/di")
    def di(name: Annotated[str, Depends(tf)]):
        return name

    with httpx.Client(
        base_url="http://testserver",
        transport=httpx.WSGITransport(app=app),  # type: ignore
    ) as client:
        resp = client.get("/?name=123")
        assert resp.text == "123"

        resp = client.get("/di?name=123")
        assert resp.text == "123"


def test_request_depends_cache():
    app = Kui()

    closed = False

    def di():
        request.state["di"] = "start"
        yield
        request.state["di"] = "end"
        nonlocal closed
        closed = True

    @auto_params
    def d(t: str = Depends(di)):
        assert request.state["di"] == "start"

    @app.router.http.get("/")
    def index(t: Annotated[str, Depends(di)]):
        d()
        assert request.state["di"] == "start"
        return b""

    with httpx.Client(
        base_url="http://testserver",
        transport=httpx.WSGITransport(app=app),  # type: ignore
    ) as client:
        resp = client.get("/")
        assert resp.text == ""
        assert closed


def test_request_depends_exception():
    app = Kui()

    closed = False

    def di():
        request.state["di"] = "start"
        try:
            yield
        except NotImplementedError:
            request.state["di"] = "end"
            nonlocal closed
            closed = True
            raise

    @app.router.http.get("/")
    def index(t: Annotated[str, Depends(di)]):
        raise NotImplementedError

    with httpx.Client(
        base_url="http://testserver",
        transport=httpx.WSGITransport(app=app),  # type: ignore
    ) as client:
        with pytest.raises(NotImplementedError):
            client.get("/")
        assert closed


def test_request_depends_cache_exception():
    app = Kui()

    closed = False

    def di():
        request.state["di"] = "start"
        try:
            yield
        except NotImplementedError:
            request.state["di"] = "end"
            nonlocal closed
            closed = True
            raise

    @auto_params
    def d(t: str = Depends(di)):
        assert request.state["di"] == "start"

    @app.router.http.get("/")
    def index(t: Annotated[str, Depends(di, cache=True)]):
        raise NotImplementedError

    with httpx.Client(
        base_url="http://testserver",
        transport=httpx.WSGITransport(app=app),  # type: ignore
    ) as client:
        with pytest.raises(NotImplementedError):
            client.get("/")
        assert closed



from __future__ import annotations

from kui.asgi import Routes
from kui.routing.extensions import FileRoutes


def test_empty_fileroutes():
    assert (Routes() + FileRoutes("kui.openapi")) == (
        Routes() + FileRoutes("kui.routing")
    )



from pathlib import Path

import httpx
import pytest
from baize.asgi import Files, Response
from pydantic import BaseModel

from kui.asgi import Kui


@pytest.mark.asyncio
async def test_pydantic_base_model():
    class Message(BaseModel):
        message: str

    app = Kui()

    @app.router.http("/message")
    async def message():
        return Message(message="Hello, World!")

    @app.router.http("/{_:any}")
    async def static_files():
        return Files(Path(__file__).absolute().parent, handle_404=Response(404))

    async with httpx.AsyncClient(
        transport=httpx.ASGITransport(app=app), base_url="http://testserver"
    ) as client:
        response = await client.get("/message")
        assert response.status_code == 200
        assert response.json() == {"message": "Hello, World!"}

        response = await client.get("/test_response_convertors.py")
        assert response.status_code == 200



from __future__ import annotations

import re

import httpx
import pytest

from kui.asgi import allow_cors


@pytest.mark.asyncio
async def test_cors():
    cors_middleware = allow_cors()

    from kui.asgi import HttpRoute, Kui

    app = Kui()

    async def homepage():
        return "homepage"

    app.router <<= HttpRoute("/", homepage) @ cors_middleware

    async with httpx.AsyncClient(
        transport=httpx.ASGITransport(app=app),
        base_url="http://testserver",
        headers={"origin": "testserver"},
    ) as client:
        resp = await client.get("/")
        assert resp.headers["access-control-allow-origin"] == "testserver"

        resp = await client.options("/")
        assert resp.headers["access-control-allow-origin"] == "testserver"


@pytest.mark.asyncio
async def test_cors_global():
    from kui.asgi import HttpRoute, Kui

    app = Kui(
        cors_config={
            "allow_origins": [
                re.compile("testserver"),
            ]
        }
    )

    async def homepage():
        return "homepage"

    app.router <<= HttpRoute("/", homepage)

    async with httpx.AsyncClient(
        transport=httpx.ASGITransport(app=app),
        base_url="http://testserver",
        headers={"origin": "testserver"},
    ) as client:
        resp = await client.get("/")
        assert resp.headers["access-control-allow-origin"] == "testserver"

        resp = await client.options("/")
        assert resp.headers["access-control-allow-origin"] == "testserver"



from __future__ import annotations

import argparse
import inspect
import os
import sys
import typing

from ..utils import import_from_string
from ..utils.inspect import get_object_filepath, get_raw_handler
from .extensions.multimethod import is_multimethod_view


def display_urls() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "application", type=str, help="Application path like: module:attr"
    )
    args = parser.parse_args()
    application = args.application

    sys.path.insert(0, os.getcwd())
    app: typing.Any = import_from_string(application)
    for path, handler in app.router.http_tree.iterator():
        print("* ", end="")
        print(path, end="")
        print(" => ", end="")

        handler = get_raw_handler(handler)

        if is_multimethod_view(handler):
            print("Is multi-method Endpoint")
            for method in handler.__methods__:
                func = get_raw_handler(getattr(handler, method.lower()))
                filepath = get_object_filepath(func)
                whitespaces = " " * (len(path) + len("* ") + len(" => "))
                print(whitespaces + "| " + method + " => ", end="")
                print(filepath + ":" + str(func.__code__.co_firstlineno))
        else:
            filepath = get_object_filepath(handler)
            print(filepath + ":" + str(inspect.getsourcelines(handler)[1]))



from __future__ import annotations

import functools
import inspect
from contextlib import AsyncExitStack, asynccontextmanager, contextmanager
from typing import Any, Callable, Dict, List, Tuple, Type, TypeVar
from typing import cast as typing_cast

from baize.datastructures import FormData
from pydantic import BaseModel

from ..parameters import (
    _convert_model_data_to_keyword_arguments,
    _create_new_signature,
    _merge_multi_value,
    _parse_depends_attrs,
    _parse_parameters_and_request_body_to_model,
    _update_docs,
    _validate_parameters_and_request_body,
    create_auto_params,
)
from ..parameters.fields import Depends
from ..utils import is_async_gen_callable, is_coroutine_callable, is_gen_callable
from .requests import http_connection, request

CallableObject = TypeVar("CallableObject", bound=Callable)


__all__ = [
    "auto_params",
]


async def call_dependencies_injection(
    depend_functions: Dict[str, Callable[..., Any]],
    depend_attrs: Dict[str, Depends],
    cache: Dict[Any, Any],
    close_soon: AsyncExitStack,
    after_response: AsyncExitStack,
) -> Dict[str, Any]:
    keyword_params: Dict[str, Any] = {}
    for name, function in depend_functions.items():
        info = depend_attrs[name]
        if info.cache and info.call in cache:
            keyword_params[name] = cache[info.call]
            continue
        if is_async_gen_callable(info.call):
            asyncgenerator = asynccontextmanager(function)()
            if inspect.isawaitable(asyncgenerator.gen):
                asyncgenerator.gen = await asyncgenerator.gen
            if info.cache:
                keyword_params[name] = await after_response.enter_async_context(
                    asyncgenerator
                )
            else:
                keyword_params[name] = await close_soon.enter_async_context(
                    asyncgenerator
                )
        elif is_coroutine_callable(info.call):
            keyword_params[name] = await function()
        elif is_gen_callable(info.call):
            generator = contextmanager(function)()
            if inspect.isawaitable(generator.gen):
                generator.gen = await generator.gen
            if info.cache:
                keyword_params[name] = after_response.enter_context(generator)
            else:
                keyword_params[name] = close_soon.enter_context(generator)
        else:
            result = function()
            if inspect.isawaitable(result):
                result = await result
            keyword_params[name] = result

        if info.cache:
            cache[info.call] = keyword_params[name]
    return keyword_params


def _create_new_callback(callback: CallableObject) -> CallableObject:
    sig = inspect.signature(callback)

    (
        parameters,
        request_body,
        exclusive_models,
        security_info,
    ) = _parse_parameters_and_request_body_to_model(sig)

    depend_attrs = _parse_depends_attrs(sig)
    depend_functions = {
        name: _create_new_callback(info.call) for name, info in depend_attrs.items()
    }

    if not (parameters or request_body or depend_attrs):
        callback_with_auto_bound_params = callback  # type: ignore
    else:

        @functools.wraps(callback)
        async def callback_with_auto_bound_params(*args, **kwargs) -> Any:
            keyword_params: Dict[str, Any] = {}

            after_response = http_connection.exit_stack
            async with AsyncExitStack() as close_soon:
                # try to call depend functions
                cache = http_connection.state.setdefault("depend_functions_cache", {})
                depends_result = await call_dependencies_injection(
                    depend_functions,
                    depend_attrs,
                    cache,
                    close_soon=close_soon,
                    after_response=after_response,
                )
                keyword_params.update(depends_result)

                data: List[Tuple[Type[BaseModel], Any]]

                try:
                    g = _validate_parameters_and_request_body(
                        parameters or {}, request_body, http_connection
                    )
                    g.send(None)
                    _body_data = await request.data()
                    if isinstance(_body_data, FormData):
                        _body_data = _merge_multi_value(_body_data.multi_items())
                    g.send(_body_data)
                except StopIteration as e:
                    data = e.value
                else:
                    raise NotImplementedError

                keyword_params.update(
                    _convert_model_data_to_keyword_arguments(data, exclusive_models)
                )

                result = callback(*args, **{**keyword_params, **kwargs})
                if inspect.isawaitable(result):
                    result = await result
                return result

        del callback_with_auto_bound_params.__wrapped__  # type: ignore

        setattr(
            callback_with_auto_bound_params, "__signature__", _create_new_signature(sig)
        )

    _update_docs(
        callback,
        callback_with_auto_bound_params,
        parameters,
        request_body,
        depend_functions,
        security_info,
    )

    return typing_cast(CallableObject, callback_with_auto_bound_params)


auto_params = create_auto_params(_create_new_callback)



from typing import Iterable, Pattern

from typing_extensions import TypedDict


class CORSConfig(TypedDict, total=False):
    allow_origins: Iterable[Pattern]
    allow_methods: Iterable[str]
    allow_headers: Iterable[str]
    expose_headers: Iterable[str]
    allow_credentials: bool
    max_age: int



from __future__ import annotations

import functools

import pytest


def test_decorator():
    from kui.asgi import Kui

    app = Kui()

    @app.router.http("/hello", name="hello")
    async def hello(): ...

    @app.router.websocket("/hello", name="hello_ws")
    async def hello_ws(): ...

    assert app.router.search("http", "/hello")[0] == {}
    assert app.router.search("websocket", "/hello")[0] == {}

    assert app.router.search("http", "/hello") != (
        app.router.search("websocket", "/hello")
    )


def test_decorator_required_method():
    from kui.asgi import Kui

    app = Kui()

    @app.router.http.get("/get")
    async def need_get(): ...

    @app.router.http.post("/post")
    async def need_post(): ...

    @app.router.http.put("/put")
    async def need_put(): ...

    @app.router.http.patch("/patch")
    async def need_patch(): ...

    @app.router.http.delete("/delete")
    async def need_delete(): ...

    assert app.router.search("http", "/get")[1].__method__ == "GET"
    assert app.router.search("http", "/post")[1].__method__ == "POST"
    assert app.router.search("http", "/put")[1].__method__ == "PUT"
    assert app.router.search("http", "/patch")[1].__method__ == "PATCH"
    assert app.router.search("http", "/delete")[1].__method__ == "DELETE"


def test_lshift():
    from kui.asgi import HttpRoute, Kui, SocketRoute

    app = Kui()

    async def hello():
        return "hello world"

    async def hello_ws(): ...

    app.router = (
        app.router
        << HttpRoute("/hello", hello, name="hello")
        << SocketRoute("/hello", hello_ws, name="hello_ws")
    )

    assert app.router.search("http", "/hello")[0] == {}
    assert app.router.search("websocket", "/hello")[0] == {}

    assert app.router.search("http", "/hello") != (
        app.router.search("websocket", "/hello")
    )


def test_url_for():
    from kui.asgi import Kui

    app = Kui()

    @app.router.http("/hello", name="hello")
    @app.router.http("/hello/{name}", name="hello-with-name")
    async def hello(request):
        return f"hello {request.path_params.get('name')}"

    assert app.router.url_for("hello") == "/hello"
    assert app.router.url_for("hello-with-name", {"name": "Aber"}) == "/hello/Aber"


def test_prefix():
    from kui.asgi import HttpRoute, Routes

    async def endpoint():
        pass

    assert [
        route.path
        for route in (
            "/auth"
            // Routes(
                HttpRoute("/login", endpoint),
                HttpRoute("/register", endpoint),
            )
        )
    ] == [
        route.path
        for route in (
            Routes(
                HttpRoute("/auth/login", endpoint),
                HttpRoute("/auth/register", endpoint),
            )
        )
    ]


def test_routes_operator():
    from kui.asgi import HttpRoute, Routes

    async def endpoint():
        pass

    routes = Routes()
    routes = routes << Routes(
        HttpRoute("/login", endpoint),
        HttpRoute("/register", endpoint),
    )
    assert routes == Routes(
        HttpRoute("/login", endpoint),
        HttpRoute("/register", endpoint),
    )

    assert (
        routes
        << Routes(
            HttpRoute("/login", endpoint),
            HttpRoute("/register", endpoint),
        )
    ) == (
        Routes(
            HttpRoute("/login", endpoint),
            HttpRoute("/register", endpoint),
        )
        + Routes(
            HttpRoute("/login", endpoint),
            HttpRoute("/register", endpoint),
        )
    )


def test_empty_path():
    from kui.asgi import Kui

    app = Kui()

    @app.router.http("")
    async def homepage():
        return "homepage"

    assert app.router.search("http", "/")[0] == {}


def test_routes_slice():
    from kui.asgi import HttpRoute, Routes

    async def endpoint():
        pass

    routes = Routes(
        HttpRoute("/login", endpoint),
        HttpRoute("/register", endpoint),
    )

    assert routes[:] == routes
    assert routes[1:] == [routes[1]]


def test_middleware_check_wraps():
    from kui.asgi import HttpRoute

    async def endpoint():
        return "hello world"

    def middleware(endpoint):
        @functools.wraps(endpoint)
        async def wrapped():
            return endpoint()

        return wrapped

    with pytest.raises(
        RuntimeError,
        match="Cannot use `@functools.wraps` on a middleware.",
    ):
        _ = HttpRoute("/hello", endpoint) @ middleware



from __future__ import annotations

import pytest

from kui.utils.importer import import_from_string, import_module


def test_import_module():
    assert import_module("sys") is None
    assert import_module("kui")


def test_import_from_string():
    assert (
        import_from_string("kui.utils.importer:import_from_string")
        is import_from_string
    )

    with pytest.raises(ValueError):
        import_from_string("kui.utils.importer")

    with pytest.raises(ImportError):
        import_from_string("kui.utils.xxxxxxxx:import_from_string")

    with pytest.raises(AttributeError):
        import_from_string("kui.utils.importer:import_from_string:foo")



from __future__ import annotations

from typing import Any, Mapping

from baize.datastructures import URL
from baize.wsgi import SmallResponse
from typing_extensions import Protocol

from .requests import request
from .responses import HttpResponse

__all__ = [
    "BaseTemplates",
    "Jinja2Templates",
]


class BaseTemplates(Protocol):
    def TemplateResponse(
        self,
        name: str,
        context: Mapping[str, Any],
        status_code: int = 200,
        headers: Mapping[str, str] | None = None,
        media_type: str | None = None,
        charset: str | None = None,
    ) -> HttpResponse:
        """
        The subclass must override this method and return
        an instance of a Response object.
        """
        raise NotImplementedError


try:
    import jinja2

    from ..templates import Jinja2TemplatesBase
except ImportError:

    class Jinja2Templates:
        def __init__(self) -> None:
            raise NotImplementedError('Install "jinja2" to use Jinja2Templates')

else:

    class _Jinja2TemplateResponse(SmallResponse):
        media_type = "text/html"

        def __init__(
            self,
            env: jinja2.Environment,
            name: str,
            context: Mapping[str, Any],
            status_code: int = 200,
            headers: Mapping[str, str] | None = None,
            media_type: str | None = None,
            charset: str | None = None,
        ):
            self.env = env
            self.template = self.env.get_template(name)
            super().__init__(context, status_code, headers, media_type, charset)

        def render(self, context: Mapping[str, Any]) -> bytes:
            text = self.template.render(context)
            return text.encode(self.charset)

    @jinja2.pass_context
    def url_for(context: dict, name: str, path_params: Mapping[str, Any]) -> URL:
        return request.url_for(name, path_params)

    class Jinja2Templates(Jinja2TemplatesBase):  # type: ignore
        """
        templates = Jinja2Templates("dir", "package:dir")

        return templates.TemplateResponse("index.html", {"request": request})
        """

        def __init__(self, *directories: str) -> None:
            super().__init__(*directories)
            self.env.globals["url_for"] = url_for

        def TemplateResponse(
            self,
            name: str,
            context: Mapping[str, Any],
            status_code: int = 200,
            headers: Mapping[str, str] | None = None,
            media_type: str | None = None,
            charset: str | None = None,
        ) -> _Jinja2TemplateResponse:
            return _Jinja2TemplateResponse(
                self.env,
                name,
                context,
                status_code=status_code,
                headers=headers,
                media_type=media_type,
                charset=charset,
            )



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Type, Union
from typing import cast as typing_cast

from baize.datastructures import ContentType
from pydantic import BaseModel
from typing_extensions import get_args, get_type_hints

from ..pydantic_compatible import get_model_fields, get_model_json_schema
from ..utils import safe_issubclass

if TYPE_CHECKING:
    from ..asgi import Kui as ASGIKui
    from ..wsgi import Kui as WSGIKui

from . import specification as spec
from .types import UploadFile


def _schema(model: Type[BaseModel]) -> spec.Schema:
    schema = get_model_json_schema(model)
    schema.pop("title")
    return typing_cast(spec.Schema, schema)


def schema_request_body(
    body: Type[BaseModel] | None, application: ASGIKui | WSGIKui
) -> Optional[spec.RequestBody]:
    if body is None:
        return None

    content_types = [
        str(v)
        for v in get_args(
            get_type_hints(application.factory_class.http.data, include_extras=True)[
                "return"
            ]
        )
        if isinstance(v, ContentType)
    ]

    for field in get_model_fields(body).values():
        if safe_issubclass(field.annotation, UploadFile):
            content_types = ["multipart/form-data"]

    return {
        "required": True,
        "content": {
            content_type: {"schema": _schema(body)} for content_type in content_types
        },
    }


def schema_response(content: Union[Type[BaseModel], spec.Schema]) -> spec.Schema:
    if isinstance(content, dict):
        return content
    else:
        return _schema(content)



from __future__ import annotations

import inspect
import operator
import typing
from dataclasses import dataclass
from functools import reduce

from typing_extensions import Self

from ..parameters import update_wrapper
from .typing import MiddlewareType, ViewType


@dataclass
class BaseRoute(typing.Generic[ViewType]):
    path: str
    endpoint: ViewType
    name: typing.Optional[str] = ""

    _auto_params: typing.ClassVar

    def extend_middlewares(self, routes: typing.Iterable[BaseRoute[ViewType]]) -> None:
        raise NotImplementedError

    def _extend_middlewares(
        self,
        middlewares: typing.Iterable[MiddlewareType],
    ) -> None:
        reduce(operator.matmul, middlewares, self)

    def __matmul__(self: Self, middleware: MiddlewareType) -> Self:
        endpoint = self.endpoint
        if hasattr(endpoint, "__methods__"):
            for method in map(str.lower, endpoint.__methods__):
                old_callback = getattr(endpoint, method)
                new_callback = middleware(old_callback)
                if getattr(new_callback, "__wrapped__", None) is old_callback:
                    raise RuntimeError("Cannot use `@functools.wraps` on a middleware.")
                if new_callback is not old_callback:
                    update_wrapper(new_callback, old_callback)
                    new_callback = self._auto_params(new_callback)
                setattr(endpoint, method, staticmethod(new_callback))
        else:
            old_callback = endpoint
            new_callback = middleware(old_callback)
            if getattr(new_callback, "__wrapped__", None) is old_callback:
                raise RuntimeError("Cannot use `@functools.wraps` on a middleware.")
            if new_callback is not old_callback:
                update_wrapper(new_callback, old_callback)
                new_callback = self._auto_params(new_callback)
            self.endpoint = new_callback  # type: ignore
        return self

    def __post_init__(self) -> None:
        assert self.path.startswith("/") or self.path == "", (
            "Route path must start with '/'"
        )
        if self.name == "":
            self.name = self.endpoint.__name__
        self.endpoint = self._auto_params(self.endpoint)


@dataclass
class HttpRoute(BaseRoute[ViewType], typing.Generic[ViewType]):
    summary: typing.Optional[str] = None
    description: typing.Optional[str] = None
    tags: typing.Optional[typing.Iterable[str]] = None

    def __post_init__(self) -> None:
        super().__post_init__()

        w: typing.Any
        if inspect.ismethod(self.endpoint):
            w = self.endpoint.__func__
        else:
            w = self.endpoint

        if self.summary:
            setattr(w, "__docs_summary__", self.summary)

        if self.description:
            setattr(w, "__docs_description__", self.description)

        if self.tags:
            setattr(w, "__docs_tags__", list(self.tags))

    def extend_middlewares(self, routes: typing.Iterable[BaseRoute[ViewType]]) -> None:
        self._extend_middlewares(getattr(routes, "_http_middlewares", []))


@dataclass
class SocketRoute(BaseRoute[ViewType], typing.Generic[ViewType]):
    def extend_middlewares(self, routes: typing.Iterable[BaseRoute[ViewType]]) -> None:
        self._extend_middlewares(getattr(routes, "_socket_middlewares", []))



from __future__ import annotations

import pytest

from kui.utils import State


@pytest.mark.asyncio
async def test_state():
    state = State({"message": "hello world"})
    with state:
        assert state.message == "hello world"
        state.like = "you"
        assert state.like == "you"
        del state.like

    async with state:
        assert state.message == "hello world"

    assert state.get("like") is None



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable

from typing_extensions import Literal

__all__ = [
    "BaseHTTPFieldInfo",
    "InPath",
    "InQuery",
    "InHeader",
    "InCookie",
    "InBody",
    "Depends",
]


@dataclass
class BaseHTTPFieldInfo:
    _in: Literal["path", "query", "header", "cookie", "body"]
    exclusive: bool = False


@dataclass
class InPath(BaseHTTPFieldInfo):
    _in: Literal["path"] = "path"


@dataclass
class InQuery(BaseHTTPFieldInfo):
    _in: Literal["query"] = "query"
    security: Any = None


@dataclass
class InHeader(BaseHTTPFieldInfo):
    _in: Literal["header"] = "header"
    security: Any = None


@dataclass
class InCookie(BaseHTTPFieldInfo):
    _in: Literal["cookie"] = "cookie"
    security: Any = None


@dataclass
class InBody(BaseHTTPFieldInfo):
    _in: Literal["body"] = "body"


@dataclass
class Depends:
    call: Callable
    cache: bool = True



from __future__ import annotations

import copy
import typing
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Sequence, TypeVar

from typing_extensions import Literal, TypedDict

if TYPE_CHECKING:
    from ..asgi import HttpRequest as ASGIHttpRequest
    from ..asgi import Kui as ASGIKui
    from ..wsgi import HttpRequest as WSGIHttpRequest
    from ..wsgi import Kui as WSGIKui

from ..exceptions import RequestValidationError
from ..parameters import _get_response_docs
from ..pydantic_compatible import DEFINITIONS_KEY
from . import specification as spec
from .extra_docs import merge_openapi_info
from .schema import schema_request_body, schema_response


class TagDetailOptional(TypedDict, total=False):
    paths: Sequence[str]


class TagDetail(TagDetailOptional):
    description: str


class OpenAPI:
    def __init__(
        self,
        info: spec.Info = {"title": "Kuí API", "version": "1.0.0"},
        security_schemes: Dict[str, spec.SecurityScheme | spec.Reference] = {},
        *,
        tags: Dict[str, TagDetail] = {},
        template_name: Literal["redoc", "swagger", "rapidoc"] = "swagger",
        template: str = "",
        reload: bool = True,
    ) -> None:
        if template == "":
            template = (
                Path(__file__).absolute().parent / "templates" / f"{template_name}.html"
            ).read_text(encoding="utf8")

        self.html_template = template
        self.reload = reload

        self.openapi = spec.OpenAPI(
            openapi="3.1.0",
            info=info,
            paths={},
            tags=[
                {"name": tag_name, "description": tag_info.get("description", "")}
                for tag_name, tag_info in tags.items()
            ],
            components={},
        )
        if security_schemes:
            components = self.openapi.setdefault("components", {})
            components["securitySchemes"] = security_schemes

        self.security_schemes: Dict[str, spec.SecurityScheme | spec.Reference] = {}

        self.path2tag: Dict[str, List[str]] = {}
        for tag_name, tag_info in tags.items():
            for path in tag_info.get("paths", []):
                self.path2tag.setdefault(path, []).append(tag_name)

    def _generate_paths(self, application: ASGIKui | WSGIKui) -> spec.Paths:
        return {
            path: openapi_path_item
            for path, openapi_path_item in (
                (path_format, self._generate_path(application, handler, path_format))
                for path_format, handler in application.router.http_tree.iterator()
            )
            if openapi_path_item
        }

    def _generate_path(
        self, application: ASGIKui | WSGIKui, view: Any, path: str
    ) -> spec.PathItem:
        """
        Generate documents under a path
        """
        if hasattr(view, "__methods__"):
            result = _clear_empty(
                {
                    method: self._generate_method(
                        application, getattr(view, method), path
                    )
                    for method in (
                        method.lower()
                        for method in typing.cast(Iterable[str], view.__methods__)
                        if method.lower() != "options"
                    )
                }
            )
        elif hasattr(view, "__method__"):
            result = _clear_empty(
                {
                    typing.cast(str, view.__method__).lower(): self._generate_method(
                        application, view, path
                    )
                }
            )
        else:
            result = {}

        return typing.cast(spec.PathItem, result)

    def _generate_method(
        self, application: ASGIKui | WSGIKui, func: Any, path: str
    ) -> spec.Operation:
        result: Dict[str, Any] = {}

        # generate summary and description
        if hasattr(func, "__docs_summary__") and isinstance(func.__docs_summary__, str):
            result["summary"] = func.__docs_summary__
        if hasattr(func, "__docs_description__") and isinstance(
            func.__docs_description__, str
        ):
            result["description"] = func.__docs_description__

        # generate security
        security_set = {}
        for security_dict in getattr(func, "__docs_security__", []):
            self.security_schemes.update(copy.deepcopy(security_dict["scheme"]))
            security_set.update(copy.deepcopy(security_dict["required"]))
        result["security"] = [{k: v} for k, v in security_set.items()]

        # generate params schema
        result["parameters"] = parameters = getattr(func, "__docs_parameters__", [])

        # generate request body schema
        request_body = schema_request_body(
            _create_model(getattr(func, "__docs_request_body__", [])), application
        )

        result["requestBody"] = request_body

        # generate responses schema
        responses: spec.Responses = {}
        __docs_responses__: List[spec.Responses] = getattr(
            func, "__docs_responses__", []
        )
        if parameters or request_body:
            handler = application.exception_middleware._lookup_exception_handler(
                RequestValidationError
            )
            if handler is None:
                raise RuntimeError
            __docs_responses__.extend(_get_response_docs(handler))

        for response_docs in __docs_responses__:
            for k, v in list(response_docs.items()):
                del response_docs[k]
                response_docs[str(k)] = v

            for response in response_docs.values():
                for media_type, media_type_value in list(
                    response.get("content", {}).items()
                ):
                    schema = schema_response(media_type_value["schema"])  # type: ignore
                    response.get("content", {})[media_type]["schema"] = schema

            need_merge_status_codes = set(responses.keys()) & set(response_docs.keys())
            if need_merge_status_codes:
                for status_code in need_merge_status_codes:
                    content = {
                        **responses[status_code].get("content", {}),
                        **response_docs[status_code].get("content", {}),
                    }
                    if content:
                        responses[status_code]["content"] = content
            else:
                responses.update(response_docs)

        result["responses"] = responses

        # set path tags
        if result and path in self.path2tag:
            result["tags"] = self.path2tag[path]

        result["tags"] = getattr(func, "__docs_tags__", []) + result.get("tags", [])

        # merge user custom operation info
        operation = typing.cast(
            spec.Operation,
            merge_openapi_info(
                _clear_empty(result), getattr(func, "__docs_extra__", {})
            ),
        )
        return operation

    def create_docs(self, request: ASGIHttpRequest | WSGIHttpRequest) -> spec.OpenAPI:
        openapi = copy.deepcopy(self.openapi)
        openapi["servers"] = [
            {
                "url": "/",
                "description": "Current server",
            },
            spec.Server(
                url="{scheme}://{address}/",
                description="Custom API Server Host",
                variables={
                    "scheme": {
                        "default": request.url.scheme,
                        "enum": ["http", "https"],
                        "description": "http or https",
                    },
                    "address": {
                        "default": request.url.netloc,
                        "description": "api server's host[:port]",
                    },
                },
            ),
        ]
        self.security_schemes = {}
        paths = copy.deepcopy(self._generate_paths(request.app))
        for path_item in paths.values():
            for operation in filter(lambda x: isinstance(x, dict), path_item.values()):
                operation = typing.cast(spec.Operation, operation)
                if "responses" not in operation:
                    operation["responses"] = {}
        components = openapi.setdefault("components", {})
        if self.security_schemes:
            components.setdefault("securitySchemes", {}).update(self.security_schemes)
        schemas = components.setdefault("schemas", {})
        schemas.update(**_pop_definitions(paths))
        openapi["paths"] = paths
        return openapi


_DictType = TypeVar("_DictType", bound=Dict)


def _clear_empty(d: _DictType) -> _DictType:
    return typing.cast(_DictType, {k: v for k, v in d.items() if v})


def _create_model(bases: List[type]) -> Optional[type]:
    if bases:
        return type("T_SchemaModel", tuple(bases), {})
    else:
        return None


def _pop_definitions(d: Dict[str, Any]) -> Dict[str, spec.Schema]:
    definitions: Dict[str, Any] = {}
    for key, value in d.items():
        if key == "schema":
            return value.pop(DEFINITIONS_KEY, {})

        if isinstance(value, dict):
            definitions.update(_pop_definitions(value))
        elif isinstance(value, (list, tuple)):
            for v in value:
                if isinstance(v, dict):
                    definitions.update(_pop_definitions(v))
        else:
            pass
    return definitions



from __future__ import annotations

import sys
from collections import deque
from typing import Any, Callable, Generic, Iterable, TypeVar

if sys.version_info >= (3, 10):  # pragma: no cover
    from typing import ParamSpec
else:  # pragma: no cover
    from typing_extensions import ParamSpec

P = ParamSpec("P")


class BackgroundTask:
    """
    Background task.
    """

    def __init__(
        self, func: Callable[P, Any], *args: P.args, **kwargs: P.kwargs
    ) -> None:
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __call__(self) -> None:
        self.func(*self.args, **self.kwargs)


class BackgroundTasks:
    def __init__(self, tasks: Iterable[BackgroundTask] | None = None):
        self.tasks: deque[BackgroundTask] = deque(tasks) if tasks else deque()

    def append(self, func: Callable[P, Any], *args: P.args, **kwargs: P.kwargs) -> None:
        task = BackgroundTask(func, *args, **kwargs)
        self.tasks.append(task)

    def __call__(self) -> None:
        for task in self.tasks:
            task()



from __future__ import annotations

from pathlib import Path

import httpx
import pytest


def test_application():
    from kui.asgi import Kui
    from kui.utils import State

    app = Kui()
    with pytest.raises(RuntimeError):
        app.state = State()

    with pytest.raises(RuntimeError):
        del app.state


@pytest.mark.asyncio
async def test_example_application():
    from example import app

    app.state.wait_time = 0.00000001

    async with httpx.AsyncClient(
        base_url="http://testserver", transport=httpx.ASGITransport(app=app)
    ) as client:
        response = await client.get("/")
        assert response.status_code == 200
        assert response.text == "Kuí"

        response = await client.get("/message")
        assert response.status_code == 200
        assert (
            response.text.replace(": ping\n\n", "")
            == "\n".join(f"id: {i}\ndata: hello\n" for i in range(5)) + "\n"
        )

        response = await client.get("/sources/README.md")
        assert response.status_code == 200
        assert response.content == (Path(".").absolute() / "README.md").read_bytes()

        response = await client.get("/sources/example")
        assert response.status_code == 404


def test_custom_application_response_converter():
    from dataclasses import asdict, dataclass
    from typing import Mapping

    from kui.asgi import HttpResponse, JSONResponse, Kui, PlainTextResponse

    @dataclass
    class Error:
        code: int = 0
        title: str = ""
        message: str = ""

    app = Kui(
        response_converters={
            Error: lambda error, status=400, headers=None: JSONResponse(
                asdict(error), status, headers
            ),
        }
    )

    @app.response_converter.register(tuple)
    @app.response_converter.register(list)
    @app.response_converter.register(dict)
    def _more_json(
        body, status: int = 200, headers: Mapping[str, str] | None = None
    ) -> HttpResponse:
        return PlainTextResponse(str(body), status, headers)

    assert isinstance(app.response_converter(Error()), JSONResponse)
    assert isinstance(app.response_converter(tuple()), PlainTextResponse)
    assert isinstance(app.response_converter(list()), PlainTextResponse)
    assert isinstance(app.response_converter(dict()), PlainTextResponse)



from __future__ import annotations

from typing import Any

from baize.datastructures import UploadFile as _UploadFile

from ..pydantic_compatible import IS_V1

if IS_V1:

    class UploadFile(_UploadFile):  # type: ignore
        """
        wrap baize UploadFile
        """

        @classmethod
        def validate(cls, v):
            if not isinstance(v, _UploadFile):
                raise TypeError(
                    f"Expected UploadFile, received {v.__class__.__name__} instead."
                )
            return v

        def __repr__(self):
            return f"File({self.filename})"

        @classmethod
        def __get_validators__(cls):
            yield cls.validate

        @classmethod
        def __modify_schema__(cls, field_schema):
            field_schema.update(type="string", format="binary")


else:
    from pydantic import GetCoreSchemaHandler, GetJsonSchemaHandler
    from pydantic.json_schema import JsonSchemaValue
    from pydantic_core.core_schema import (
        CoreSchema,
        any_schema,
        no_info_after_validator_function,
    )

    class UploadFile(_UploadFile):  # type: ignore
        """
        wrap baize UploadFile
        """

        @classmethod
        def validate(cls, v):
            if not isinstance(v, _UploadFile):
                raise ValueError(
                    f"Expected UploadFile, received {v.__class__.__name__} instead."
                )
            return v

        def __repr__(self):
            return f"File({self.filename})"

        @classmethod
        def __get_pydantic_core_schema__(
            cls, _source_type: Any, _handler: GetCoreSchemaHandler
        ) -> CoreSchema:
            return no_info_after_validator_function(cls.validate, any_schema())

        @classmethod
        def __get_pydantic_json_schema__(
            cls, schema: CoreSchema, handler: GetJsonSchemaHandler
        ) -> JsonSchemaValue:
            json_schema = handler(schema)
            json_schema.update(type="string", format="binary")
            return json_schema



from __future__ import annotations

import copy
import dataclasses
import functools
import sys
from pathlib import PurePath
from types import AsyncGeneratorType
from typing import (
    Any,
    Callable,
    Iterable,
    List,
    Mapping,
    NoReturn,
    Optional,
    Type,
    TypeVar,
)

from baize.asgi import Files, Hosts, Pages, Subpaths
from baize.asgi import Router as BaizeRouter
from baize.datastructures import URL
from baize.typing import Receive, Scope, Send
from pydantic import BaseModel
from typing_extensions import Literal

from ..cors import CORSConfig
from ..responses import create_json_encoder
from ..routing import AsyncViewType, BaseRoute, MiddlewareType, NoMatchFound
from ..utils import ImmutableAttribute, State
from ..utils.contextvars import context_setter
from .cors import allow_cors
from .exceptions import ErrorHandlerType, ExceptionMiddleware, HTTPException
from .lifespan import Lifespan, LifespanCallback
from .requests import (
    HttpRequest,
    WebSocket,
    http_connection_var,
    request_var,
    websocket_var,
)
from .responses import (
    FileResponse,
    HttpResponse,
    JSONResponse,
    PlainTextResponse,
    RedirectResponse,
    SendEventResponse,
)
from .routing import Router
from .templates import BaseTemplates

LifespanCallbackTypeVar = TypeVar("LifespanCallbackTypeVar", bound=LifespanCallback)


@dataclasses.dataclass
class FactoryClass:
    http: Type[HttpRequest] = HttpRequest
    websocket: Type[WebSocket] = WebSocket


class Kui:
    state: ImmutableAttribute[State] = ImmutableAttribute()

    def __init__(
        self,
        *,
        templates: Optional[BaseTemplates] = None,
        on_startup: List[LifespanCallback] = [],
        on_shutdown: List[LifespanCallback] = [],
        routes: Iterable[BaseRoute] = [],
        http_middlewares: List[MiddlewareType] = [],
        socket_middlewares: List[MiddlewareType] = [],
        exception_handlers: Mapping[int | Type[BaseException], ErrorHandlerType] = {},
        cors_config: Optional[CORSConfig] = None,
        factory_class: FactoryClass = FactoryClass(),
        response_converters: Mapping[type, Callable[..., HttpResponse]] = {},
        json_encoder: Mapping[type, Callable[[Any], Any]] = {},
    ) -> None:
        self.should_exit = False

        self.state = State()
        self.response_converter = create_response_converter(response_converters)
        self.json_encoder = create_json_encoder(*json_encoder.items())
        self.factory_class = factory_class
        self.templates = templates
        self.lifespan = Lifespan(copy.copy(on_startup), copy.copy(on_shutdown))

        http_middlewares = [*http_middlewares]

        self.exception_middleware = ExceptionMiddleware(exception_handlers)
        http_middlewares.append(self.exception_middleware)

        if cors_config is not None:
            http_middlewares.append(allow_cors(**cors_config))

        self.router = Router(routes, http_middlewares, socket_middlewares)

    def add_exception_handler(
        self, exc_class_or_status_code: int | Type[Exception], handler: ErrorHandlerType
    ) -> None:
        self.exception_middleware.add_exception_handler(
            exc_class_or_status_code, handler
        )

    def exception_handler(
        self, exc_class_or_status_code: int | Type[Exception]
    ) -> Callable[[ErrorHandlerType], ErrorHandlerType]:
        def decorator(func: ErrorHandlerType) -> ErrorHandlerType:
            self.add_exception_handler(exc_class_or_status_code, func)
            return func

        return decorator

    def on_startup(self, func: LifespanCallbackTypeVar) -> LifespanCallbackTypeVar:
        self.lifespan.on_startup.append(func)
        return func

    def on_shutdown(self, func: LifespanCallbackTypeVar) -> LifespanCallbackTypeVar:
        self.lifespan.on_shutdown.append(func)
        return func

    async def app(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope_type: Literal["lifespan", "http", "websocket"] = scope["type"]
        return await getattr(self, scope_type)(scope, receive, send)

    async def http(self, scope: Scope, receive: Receive, send: Send) -> None:
        request = self.factory_class.http(scope, receive, send)
        with (
            context_setter(http_connection_var, request),
            context_setter(request_var, request),
        ):
            try:
                try:
                    path_params, handler = self.router.search("http", request["path"])
                    request["path_params"] = path_params
                    response = await handler()
                except NoMatchFound:
                    http_exception: HTTPException[None] = HTTPException(404)
                    error_handler = self.exception_middleware.lookup_handler(
                        http_exception
                    )
                    if error_handler is None:
                        raise RuntimeError(
                            "No exception handler found for HTTPException(404)"
                        )
                    response = await error_handler(http_exception)

                if isinstance(response, tuple):
                    response = self.response_converter(*response)
                else:
                    response = self.response_converter(response)

                return await response(scope, receive, send)
            finally:
                try:
                    await request.exit_stack.__aexit__(*sys.exc_info())
                    await request.background_tasks()
                finally:
                    await request.close()

    async def websocket(self, scope: Scope, receive: Receive, send: Send) -> None:
        websocket = self.factory_class.websocket(scope, receive, send)
        with (
            context_setter(http_connection_var, websocket),
            context_setter(websocket_var, websocket),
        ):
            try:
                try:
                    path_params, handler = self.router.search(
                        "websocket", websocket["path"]
                    )
                    websocket["path_params"] = path_params
                except NoMatchFound:
                    return await websocket.close(1001)
                else:
                    return await handler()
            finally:
                await websocket.exit_stack.__aexit__(*sys.exc_info())
                await websocket.background_tasks()

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope["app"] = self

        await self.app(scope, receive, send)


def create_response_converter(
    converters: Mapping[type, Callable[..., HttpResponse]],
) -> functools._SingleDispatchCallable[HttpResponse]:
    """
    Create a converter for convert response.
    """

    @functools.singledispatch
    def response_converter(*args: Any) -> HttpResponse:
        raise TypeError(
            f"Cannot find response_converter handler for this type: {type(args[0])}"
        )

    def _none(ret: Type[None]) -> NoReturn:
        raise TypeError(
            "Get 'None'. Maybe you need to add a return statement to the function."
        )

    response_converter.register(type(None), _none)
    response_converter.register(HttpResponse, lambda x: x)
    response_converter.register(Files, lambda x: x)
    response_converter.register(Pages, lambda x: x)
    response_converter.register(Subpaths, lambda x: x)
    response_converter.register(BaizeRouter, lambda x: x)
    response_converter.register(Hosts, lambda x: x)
    response_converter.register(dict, JSONResponse)
    response_converter.register(list, JSONResponse)
    response_converter.register(tuple, JSONResponse)
    response_converter.register(bytes, PlainTextResponse)
    response_converter.register(str, PlainTextResponse)
    # Because of we can't call `await anext(ag)` in the `response_converter` function,
    # so we can't convert `AsyncGenerator` to `SendEventResponse | StreamResponse`.
    response_converter.register(AsyncGeneratorType, SendEventResponse)
    response_converter.register(
        PurePath,
        lambda filepath, download_name=None: FileResponse(
            str(filepath), download_name=download_name
        ),
    )
    response_converter.register(URL, RedirectResponse)
    response_converter.register(BaseModel, JSONResponse)

    for type_, converter in converters.items():
        response_converter.register(type_, converter)

    return response_converter



from __future__ import annotations

from typing import List, Union

try:
    import jinja2
except ImportError:
    pass
else:

    class Jinja2TemplatesBase:
        """
        templates = Jinja2Templates("dir", "package:dir")

        return templates.TemplateResponse("index.html", {"request": request})
        """

        def __init__(self, *directories: str) -> None:
            self.env = self.get_env(self.get_loaders(*directories))

        def get_loaders(self, *directories: str) -> jinja2.BaseLoader:
            templates_loaders: List[
                Union[jinja2.FileSystemLoader, jinja2.PackageLoader]
            ] = []
            for directory in directories:
                if ":" in directory:
                    package_name, package_path = directory.split(":", maxsplit=1)
                    templates_loaders.append(
                        jinja2.PackageLoader(package_name, package_path)
                    )
                else:
                    templates_loaders.append(jinja2.FileSystemLoader(directory))
            return jinja2.ChoiceLoader(templates_loaders)

        def get_env(self, loader: jinja2.BaseLoader) -> jinja2.Environment:
            return jinja2.Environment(loader=loader, autoescape=True)



import json
from http import HTTPStatus
from typing import Any, List, Tuple

import httpx
import pytest
from pydantic import BaseModel
from typing_extensions import Annotated

from kui.asgi import (
    Depends,
    Header,
    HTMLResponse,
    HttpRoute,
    HttpView,
    JSONResponse,
    Kui,
    OpenAPI,
    Path,
    Routes,
    api_key_auth_dependency,
    basic_auth,
    bearer_auth,
    required_method,
)


@pytest.mark.asyncio
async def test_openapi_page():
    app = Kui()
    openapi = OpenAPI()
    app.router <<= Routes("/docs" // openapi.routes, namespace="docs")

    @app.router.http.get("/hello")
    async def hello() -> Annotated[Any, JSONResponse[200, {}, List[str]]]:
        """
        hello
        """
        pass

    class Username(BaseModel):
        name: str

    @app.router.http.get("/path/{name}")
    async def path(name: Annotated[str, Path(...)]):
        pass

    @app.router.http("/http-view")
    class HTTPClass(HttpView):
        @classmethod
        async def get(cls) -> Annotated[Any, HTMLResponse[200]]:
            """
            ...

            ......
            """

        @classmethod
        async def post(cls) -> Annotated[Any, JSONResponse[201, {}, Username]]:
            """
            ...

            ......
            """

        @classmethod
        async def delete(
            cls,
        ) -> Annotated[Any, {"204": {"description": HTTPStatus(204).description}}]:
            """
            ...

            ......
            """

    def just_middleware(endpoint):
        async def wrapper(
            authorization: Annotated[str, Header(..., description="JWT Token")],
        ) -> Annotated[Any, {"401": {"description": HTTPStatus(401).description}}]:
            return await endpoint()

        return wrapper

    middleware_routes = "/middleware" // Routes(
        HttpRoute("/path/{name}", path),
        HttpRoute("/http-view", HTTPClass),
        http_middlewares=[just_middleware],
        namespace="middleware",
    )

    app.router <<= middleware_routes

    async with httpx.AsyncClient(
        transport=httpx.ASGITransport(app=app), base_url="http://testserver"
    ) as client:
        response = await client.get("/docs", follow_redirects=False)
        assert response.status_code == 307
        assert response.headers["location"] == "http://testserver/docs/"

        response = await client.get("/docs/")
        assert response.status_code == 200

        response = await client.get("/docs/json")
        assert response.status_code == 200
        assert len(response.headers["hash"]) == 32

        openapi_docs_text = response.text
        assert json.loads(openapi_docs_text) == {
            "openapi": "3.1.0",
            "info": {"title": "Kuí API", "version": "1.0.0"},
            "paths": {
                "/hello": {
                    "get": {
                        "summary": "hello",
                        "responses": {
                            "200": {
                                "description": "Request fulfilled, document follows",
                                "headers": {},
                                "content": {
                                    "application/json": {
                                        "schema": {
                                            "type": "array",
                                            "items": {"type": "string"},
                                        }
                                    }
                                },
                            }
                        },
                    }
                },
                "/http-view": {
                    "get": {
                        "summary": "...",
                        "description": "......",
                        "responses": {
                            "200": {
                                "description": "Request fulfilled, document follows",
                                "content": {
                                    "text/html": {"schema": {"type": "string"}}
                                },
                                "headers": {},
                            }
                        },
                    },
                    "post": {
                        "summary": "...",
                        "description": "......",
                        "responses": {
                            "201": {
                                "description": "Document created, URL follows",
                                "headers": {},
                                "content": {
                                    "application/json": {
                                        "schema": {
                                            "type": "object",
                                            "properties": {
                                                "name": {
                                                    "title": "Name",
                                                    "type": "string",
                                                }
                                            },
                                            "required": ["name"],
                                        }
                                    }
                                },
                            }
                        },
                    },
                    "delete": {
                        "summary": "...",
                        "description": "......",
                        "responses": {
                            "204": {"description": "Request fulfilled, nothing follows"}
                        },
                    },
                },
                "/path/{name}": {
                    "get": {
                        "parameters": [
                            {
                                "in": "path",
                                "name": "name",
                                "description": "",
                                "required": True,
                                "schema": {"title": "Name", "type": "string"},
                                "deprecated": False,
                            }
                        ],
                        "responses": {
                            "422": {
                                "description": "",
                                "headers": {},
                                "content": {
                                    "application/json": {
                                        "schema": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "loc": {
                                                        "title": "Location",
                                                        "description": "error field",
                                                        "type": "array",
                                                        "items": {"type": "string"},
                                                    },
                                                    "type": {
                                                        "title": "Type",
                                                        "description": "error type",
                                                        "type": "string",
                                                    },
                                                    "msg": {
                                                        "title": "Message",
                                                        "description": "error message",
                                                        "type": "string",
                                                    },
                                                    "ctx": {
                                                        "title": "Context",
                                                        "description": "error context",
                                                        "type": "string",
                                                    },
                                                    "in": {
                                                        "title": "In",
                                                        "type": "string",
                                                        "enum": [
                                                            "path",
                                                            "query",
                                                            "header",
                                                            "cookie",
                                                            "body",
                                                        ],
                                                    },
                                                },
                                                "required": ["loc", "type", "msg"],
                                            },
                                        }
                                    }
                                },
                            }
                        },
                    }
                },
                "/middleware/http-view": {
                    "get": {
                        "summary": "...",
                        "description": "......",
                        "parameters": [
                            {
                                "in": "header",
                                "name": "authorization",
                                "description": "JWT Token",
                                "required": True,
                                "schema": {"title": "Authorization", "type": "string"},
                                "deprecated": False,
                            }
                        ],
                        "responses": {
                            "200": {
                                "description": "Request fulfilled, document follows",
                                "content": {
                                    "text/html": {"schema": {"type": "string"}}
                                },
                                "headers": {},
                            },
                            "401": {
                                "description": "No permission -- see authorization schemes"
                            },
                            "422": {
                                "description": "",
                                "headers": {},
                                "content": {
                                    "application/json": {
                                        "schema": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "loc": {
                                                        "title": "Location",
                                                        "description": "error field",
                                                        "type": "array",
                                                        "items": {"type": "string"},
                                                    },
                                                    "type": {
                                                        "title": "Type",
                                                        "description": "error type",
                                                        "type": "string",
                                                    },
                                                    "msg": {
                                                        "title": "Message",
                                                        "description": "error message",
                                                        "type": "string",
                                                    },
                                                    "ctx": {
                                                        "title": "Context",
                                                        "description": "error context",
                                                        "type": "string",
                                                    },
                                                    "in": {
                                                        "title": "In",
                                                        "type": "string",
                                                        "enum": [
                                                            "path",
                                                            "query",
                                                            "header",
                                                            "cookie",
                                                            "body",
                                                        ],
                                                    },
                                                },
                                                "required": ["loc", "type", "msg"],
                                            },
                                        }
                                    }
                                },
                            },
                        },
                    },
                    "post": {
                        "summary": "...",
                        "description": "......",
                        "parameters": [
                            {
                                "in": "header",
                                "name": "authorization",
                                "description": "JWT Token",
                                "required": True,
                                "schema": {"title": "Authorization", "type": "string"},
                                "deprecated": False,
                            }
                        ],
                        "responses": {
                            "201": {
                                "description": "Document created, URL follows",
                                "headers": {},
                                "content": {
                                    "application/json": {
                                        "schema": {
                                            "type": "object",
                                            "properties": {
                                                "name": {
                                                    "title": "Name",
                                                    "type": "string",
                                                }
                                            },
                                            "required": ["name"],
                                        }
                                    }
                                },
                            },
                            "401": {
                                "description": "No permission -- see authorization schemes"
                            },
                            "422": {
                                "description": "",
                                "headers": {},
                                "content": {
                                    "application/json": {
                                        "schema": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "loc": {
                                                        "title": "Location",
                                                        "description": "error field",
                                                        "type": "array",
                                                        "items": {"type": "string"},
                                                    },
                                                    "type": {
                                                        "title": "Type",
                                                        "description": "error type",
                                                        "type": "string",
                                                    },
                                                    "msg": {
                                                        "title": "Message",
                                                        "description": "error message",
                                                        "type": "string",
                                                    },
                                                    "ctx": {
                                                        "title": "Context",
                                                        "description": "error context",
                                                        "type": "string",
                                                    },
                                                    "in": {
                                                        "title": "In",
                                                        "type": "string",
                                                        "enum": [
                                                            "path",
                                                            "query",
                                                            "header",
                                                            "cookie",
                                                            "body",
                                                        ],
                                                    },
                                                },
                                                "required": ["loc", "type", "msg"],
                                            },
                                        }
                                    }
                                },
                            },
                        },
                    },
                    "delete": {
                        "summary": "...",
                        "description": "......",
                        "parameters": [
                            {
                                "in": "header",
                                "name": "authorization",
                                "description": "JWT Token",
                                "required": True,
                                "schema": {"title": "Authorization", "type": "string"},
                                "deprecated": False,
                            }
                        ],
                        "responses": {
                            "204": {
                                "description": "Request fulfilled, nothing follows"
                            },
                            "401": {
                                "description": "No permission -- see authorization schemes"
                            },
                            "422": {
                                "description": "",
                                "headers": {},
                                "content": {
                                    "application/json": {
                                        "schema": {
                                            "type": "array",
                                            "items": {
                                                "type": "object",
                                                "properties": {
                                                    "loc": {
                                                        "title": "Location",
                                                        "description": "error field",
                                                        "type": "array",
                                                        "items": {"type": "string"},
                                                    },
                                                    "type": {
                                                        "title": "Type",
                                                        "description": "error type",
                                                        "type": "string",
                                                    },
                                                    "msg": {
                                                        "title": "Message",
                                                        "description": "error message",
                                                        "type": "string",
                                                    },
                                                    "ctx": {
                                                        "title": "Context",
                                                        "description": "error context",
                                                        "type": "string",
                                                    },
                                                    "in": {
                                                        "title": "In",
                                                        "type": "string",
                                                        "enum": [
                                                            "path",
                                                            "query",
                                                            "header",
                                                            "cookie",
                                                            "body",
                                                        ],
                                                    },
                                                },
                                                "required": ["loc", "type", "msg"],
                                            },
                                        }
                                    }
                                },
                            },
                        },
                    },
                },
            },
            "tags": [],
            "components": {"schemas": {}},
            "servers": [
                {"url": "/", "description": "Current server"},
                {
                    "url": "{scheme}://{address}/",
                    "description": "Custom API Server Host",
                    "variables": {
                        "scheme": {
                            "default": "http",
                            "enum": ["http", "https"],
                            "description": "http or https",
                        },
                        "address": {
                            "default": "testserver",
                            "description": "api server's host[:port]",
                        },
                    },
                },
            ],
        }, openapi_docs_text


def test_openapi_single_function_summary_and_description():
    app = Kui()
    openapi = OpenAPI()
    app.router <<= "/docs" // openapi.routes

    @app.router.http.get("/0", name=None, summary="Summary", description="Description")
    async def _():
        return ""

    @app.router.http.get("/1", name=None, summary="Summary")
    async def _():
        return ""

    @app.router.http.get("/2", name=None, summary="Summary")
    async def _():
        """
        Description
        """
        return ""

    @app.router.http.get("/3", name=None)
    async def _():
        """
        Summary

        Description
        """
        return ""

    assert openapi._generate_path(app, app.router.search("http", "/0")[1], "/") == {
        "get": {"summary": "Summary", "description": "Description"}
    }
    assert openapi._generate_path(app, app.router.search("http", "/1")[1], "/") == {
        "get": {"summary": "Summary"}
    }
    assert openapi._generate_path(app, app.router.search("http", "/2")[1], "/") == {
        "get": {"summary": "Summary"}
    }
    assert openapi._generate_path(app, app.router.search("http", "/3")[1], "/") == {
        "get": {"summary": "Summary", "description": "Description"}
    }


def test_openapi_single_function_tags():
    app = Kui()
    openapi = OpenAPI()
    app.router <<= "/docs" // openapi.routes

    @app.router.http.get("/", name=None, tags=["tag0"])
    async def homepage():
        return ""

    assert openapi._generate_path(app, app.router.search("http", "/")[1], "/") == (
        {"get": {"tags": ["tag0"]}}
    )


def test_openapi_routes_tags():
    app = Kui()
    openapi = OpenAPI()
    app.router <<= "/docs" // openapi.routes

    async def homepage():
        return ""

    app.router <<= Routes(
        HttpRoute("/", homepage) @ required_method("GET"), tags=["tag0"]
    )

    assert openapi._generate_path(app, app.router.search("http", "/")[1], "/") == (
        {"get": {"tags": ["tag0"]}}
    )


@pytest.mark.asyncio
async def test_openapi_security():
    app = Kui()
    openapi = OpenAPI()
    app.router <<= "/docs" // openapi.routes

    @app.router.http.get("/basic")
    async def basic(user_and_password: Annotated[Tuple[str, str], Depends(basic_auth)]):
        return ", ".join(user_and_password)

    @app.router.http.get("/bearer")
    async def bearer(token: Annotated[str, Depends(bearer_auth)]):
        return token

    @app.router.http.get("/api-key")
    async def api_key(
        api_key: Annotated[str, Depends(api_key_auth_dependency("api-key"))],
    ):
        return api_key

    async with httpx.AsyncClient(
        base_url="http://testserver", transport=httpx.ASGITransport(app=app)
    ) as client:
        resp = await client.get("/docs/json")
        openapi_json = resp.json()

        assert openapi_json == {
            "openapi": "3.1.0",
            "info": {"title": "Kuí API", "version": "1.0.0"},
            "paths": {
                "/basic": {
                    "get": {
                        "security": [{"BasicAuth": []}],
                        "responses": {
                            "401": {
                                "description": "No permission -- see authorization schemes",
                                "headers": {
                                    "WWW-Authenticate": {
                                        "description": "Basic authentication",
                                        "schema": {"type": "string"},
                                    }
                                },
                            }
                        },
                    }
                },
                "/bearer": {
                    "get": {
                        "security": [{"BearerAuth": []}],
                        "responses": {
                            "401": {
                                "description": "No permission -- see authorization schemes",
                                "headers": {
                                    "WWW-Authenticate": {
                                        "description": "Bearer token",
                                        "schema": {"type": "string"},
                                    }
                                },
                            }
                        },
                    }
                },
                "/api-key": {
                    "get": {
                        "security": [{"ApiKeyAuth": []}],
                        "responses": {
                            "401": {
                                "description": "No permission -- see authorization schemes"
                            }
                        },
                    }
                },
            },
            "tags": [],
            "components": {
                "securitySchemes": {
                    "BasicAuth": {"type": "http", "scheme": "basic"},
                    "BearerAuth": {"type": "http", "scheme": "bearer"},
                    "ApiKeyAuth": {"type": "apiKey", "name": "api-key", "in": "header"},
                },
                "schemas": {},
            },
            "servers": [
                {"url": "/", "description": "Current server"},
                {
                    "url": "{scheme}://{address}/",
                    "description": "Custom API Server Host",
                    "variables": {
                        "scheme": {
                            "default": "http",
                            "enum": ["http", "https"],
                            "description": "http or https",
                        },
                        "address": {
                            "default": "testserver",
                            "description": "api server's host[:port]",
                        },
                    },
                },
            ],
        }


def test_openapi_depend_response():
    app = Kui()
    openapi = OpenAPI()
    app.router <<= "/docs" // openapi.routes

    async def get_current_user(
        authorization: Annotated[str, Header(..., description="JWT Token")],
    ) -> Annotated[Any, {"401": {"description": HTTPStatus(401).description}}]:
        pass

    @app.router.http.get("/")
    async def homepage(user: Annotated[Any, Depends(get_current_user)]) -> Any:
        pass

    assert openapi._generate_path(app, app.router.search("http", "/")[1], "/") == (
        {
            "get": {
                "parameters": [
                    {
                        "in": "header",
                        "name": "authorization",
                        "description": "JWT Token",
                        "required": True,
                        "schema": {"title": "Authorization", "type": "string"},
                        "deprecated": False,
                    }
                ],
                "responses": {
                    "422": {
                        "description": "",
                        "headers": {},
                        "content": {
                            "application/json": {
                                "schema": {
                                    "type": "array",
                                    "items": {
                                        "type": "object",
                                        "properties": {
                                            "loc": {
                                                "title": "Location",
                                                "description": "error field",
                                                "type": "array",
                                                "items": {"type": "string"},
                                            },
                                            "type": {
                                                "title": "Type",
                                                "description": "error type",
                                                "type": "string",
                                            },
                                            "msg": {
                                                "title": "Message",
                                                "description": "error message",
                                                "type": "string",
                                            },
                                            "ctx": {
                                                "title": "Context",
                                                "description": "error context",
                                                "type": "string",
                                            },
                                            "in": {
                                                "title": "In",
                                                "type": "string",
                                                "enum": [
                                                    "path",
                                                    "query",
                                                    "header",
                                                    "cookie",
                                                    "body",
                                                ],
                                            },
                                        },
                                        "required": ["loc", "type", "msg"],
                                    },
                                }
                            }
                        },
                    },
                    "401": {
                        "description": "No permission -- see authorization schemes",
                    },
                },
            }
        }
    )


def test_openapi_multi_type_responses():
    app = Kui()
    openapi = OpenAPI()
    app.router <<= "/docs" // openapi.routes

    class Message(BaseModel):
        message: str

    @app.router.http.get("/")
    async def homepage() -> Annotated[
        Any, JSONResponse[200, {}, Message], HTMLResponse[200]
    ]:
        pass

    assert openapi._generate_path(app, app.router.search("http", "/")[1], "/") == {
        "get": {
            "responses": {
                "200": {
                    "description": "Request fulfilled, document follows",
                    "headers": {},
                    "content": {
                        "application/json": {
                            "schema": {
                                "type": "object",
                                "properties": {
                                    "message": {"title": "Message", "type": "string"}
                                },
                                "required": ["message"],
                            }
                        },
                        "text/html": {"schema": {"type": "string"}},
                    },
                }
            }
        }
    }


@pytest.mark.asyncio
async def test_openapi_components_schemas():
    app = Kui()
    openapi = OpenAPI()
    app.router <<= Routes("/docs" // openapi.routes, namespace="docs")

    class M(BaseModel):
        message: str

    class N(BaseModel):
        m: M

    @app.router.http.get("/hello")
    async def hello() -> Annotated[Any, JSONResponse[200, {}, N]]:
        """
        hello
        """
        pass

    async with httpx.AsyncClient(
        transport=httpx.ASGITransport(app=app), base_url="http://testserver"
    ) as client:
        response = await client.get("/docs/json")
        assert response.status_code == 200
        assert len(response.headers["hash"]) == 32

        openapi_docs_text = response.text
        assert json.loads(openapi_docs_text) == {
            "openapi": "3.1.0",
            "info": {"title": "Kuí API", "version": "1.0.0"},
            "paths": {
                "/hello": {
                    "get": {
                        "summary": "hello",
                        "responses": {
                            "200": {
                                "description": "Request fulfilled, document follows",
                                "headers": {},
                                "content": {
                                    "application/json": {
                                        "schema": {
                                            "properties": {
                                                "m": {"$ref": "#/components/schemas/M"}
                                            },
                                            "required": ["m"],
                                            "type": "object",
                                        }
                                    }
                                },
                            }
                        },
                    }
                }
            },
            "tags": [],
            "components": {
                "schemas": {
                    "M": {
                        "properties": {
                            "message": {"title": "Message", "type": "string"}
                        },
                        "required": ["message"],
                        "title": "M",
                        "type": "object",
                    }
                }
            },
            "servers": [
                {"url": "/", "description": "Current server"},
                {
                    "url": "{scheme}://{address}/",
                    "description": "Custom API Server Host",
                    "variables": {
                        "scheme": {
                            "default": "http",
                            "enum": ["http", "https"],
                            "description": "http or https",
                        },
                        "address": {
                            "default": "testserver",
                            "description": "api server's host[:port]",
                        },
                    },
                },
            ],
        }, openapi_docs_text
